from __future__ import division

"""Class definitions."""

import os, warnings
from os.path import join
import numpy as np
import pandas as pd
from pandas import DataFrame, Series, Index
import six
import abc
from copy import deepcopy
from functools import reduce
from nltools.data import Adjacency, design_matrix
from nltools.stats import downsample, upsample, transform_pairwise, regress
from nltools.utils import set_decomposition_algorithm
from sklearn.metrics.pairwise import pairwise_distances, cosine_similarity
from sklearn.utils import check_random_state
from sklearn.linear_model import LinearRegression

from feat.utils import (
    read_feat,
    read_affectiva,
    read_facet,
    read_openface,
    wavelet,
    calc_hist_auc,
    load_h5,
    get_resource_path,
)
from feat.plotting import plot_face, draw_lineface, draw_muscles
from nilearn.signal import clean
from scipy.signal import convolve
from scipy.stats import ttest_1samp


class FexSeries(Series):
    """
    This is a sub-class of pandas series. While not having additional methods
    of it's own required to retain normal slicing functionality for the
    Fex class, i.e. how slicing is typically handled in pandas.
    All methods should be called on Fex below.
    """

    _metadata = [
        "au_columns",
        "emotion_columns",
        "facebox_columns",
        "landmark_columns",
        "facepose_columns",
        "gaze_columns",
        "time_columns",
        "design_columns",
        "fex_columns",
        "filename",
        "sampling_freq",
        "features",
        "sessions",
        "detector",
    ]

    def __init__(self, *args, **kwargs):
        ### Columns ###
        self.au_columns = kwargs.pop("au_columns", None)
        self.emotion_columns = kwargs.pop("emotion_columns", None)
        self.facebox_columns = kwargs.pop("facebox_columns", None)
        self.landmark_columns = kwargs.pop("landmark_columns", None)
        self.facepose_columns = kwargs.pop("facepose_columns", None)
        self.gaze_columns = kwargs.pop("gaze_columns", None)
        self.time_columns = kwargs.pop("time_columns", None)
        self.design_columns = kwargs.pop("design_columns", None)

        ### Meta data ###
        self.filename = kwargs.pop("filename", None)
        self.sampling_freq = kwargs.pop("sampling_freq", None)
        self.detector = kwargs.pop("detector", None)
        self.features = kwargs.pop("features", None)
        self.sessions = kwargs.pop("sessions", None)
        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return FexSeries

    @property
    def _constructor_expanddim(self):
        return Fex

    def __finalize__(self, other, method=None, **kwargs):
        """ propagate metadata from other to self """
        # NOTE: backported from pandas master (upcoming v0.13)
        for name in self._metadata:
            object.__setattr__(self, name, getattr(other, name, None))
        return self

    def aus(self):
        """Returns the Action Units data

        Returns:
            DataFrame: Action Units data
        """
        return self[self.au_columns]

    def emotions(self):
        """Returns the emotion data

        Returns:
            DataFrame: emotion data
        """
        return self[self.emotion_columns]

    def landmark(self):
        """Returns the landmark data

        Returns:
            DataFrame: landmark data
        """
        return self[self.landmark_columns]

    def input(self):
        """Returns input column as string

        Returns:
            string: path to input image
        """
        return self["input"]

    def landmark_x(self):
        """Returns the x landmarks.

        Returns:
            DataFrame: x landmarks.
        """
        ######## TODO: NATSORT columns before returning #######
        x_cols = [col for col in self.landmark_columns if "x" in col]
        return self[x_cols]

    def landmark_y(self):
        """Returns the y landmarks.

        Returns:
            DataFrame: y landmarks.
        """
        y_cols = [col for col in self.landmark_columns if "y" in col]
        return self[y_cols]

    def facebox(self):
        """Returns the facebox data

        Returns:
            DataFrame: facebox data
        """
        return self[self.facebox_columns]

    def time(self):
        """Returns the time data

        Returns:
            DataFrame: time data
        """
        return self[self.time_columns]

    def design(self):
        """Returns the design data

        Returns:
            DataFrame: time data
        """
        return self[self.design_columns]

    def info(self):
        """Print class meta data."""
        attr_list = []
        for name in self._metadata:
            attr_list.append(name + ": " + str(getattr(self, name, None)) + "\n")
        print(f"{self.__class__}\n" + "".join(attr_list))


class Fex(DataFrame):
    """Fex is a class to represent facial expression (Fex) data. It is essentially
        an enhanced pandas df, with extra attributes and methods. Methods
        always return a new design matrix instance.

    Args:
        filename: (str, optional) path to file
        detector: (str, optional) name of software used to extract Fex. (Feat, FACET, OpenFace, or Affectiva)
        sampling_freq (float, optional): sampling rate of each row in Hz;
                                         defaults to None
        features (pd.Dataframe, optional): features that correspond to each
                                          Fex row
        sessions: Unique values indicating rows associated with a specific
                  session (e.g., trial, subject, etc). Must be a 1D array of
                  n_samples elements; defaults to None
    """

    # __metaclass__  = abc.ABCMeta

    # Need to specify attributes for pandas.
    _metadata = [
        "au_columns",
        "emotion_columns",
        "facebox_columns",
        "landmark_columns",
        "facepose_columns",
        "gaze_columns",
        "time_columns",
        "design_columns",
        "fex_columns",
        "filename",
        "sampling_freq",
        "features",
        "sessions",
        "detector",
    ]

    def __finalize__(self, other, method=None, **kwargs):
        """propagate metadata from other to self """
        self = super().__finalize__(other, method=method, **kwargs)
        # merge operation: using metadata of the left object
        if method == "merge":
            for name in self._metadata:
                print("self", name, self.au_columns, other.left.au_columns)
                object.__setattr__(self, name, getattr(other.left, name, None))
        # concat operation: using metadata of the first object
        elif method == "concat":
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other.objs[0], name, None))
        return self

    def __init__(self, *args, **kwargs):
        ### Columns ###
        self.au_columns = kwargs.pop("au_columns", None)
        self.emotion_columns = kwargs.pop("emotion_columns", None)
        self.facebox_columns = kwargs.pop("facebox_columns", None)
        self.landmark_columns = kwargs.pop("landmark_columns", None)
        self.facepose_columns = kwargs.pop("facepose_columns", None)
        self.gaze_columns = kwargs.pop("gaze_columns", None)
        self.time_columns = kwargs.pop("time_columns", None)
        self.design_columns = kwargs.pop("design_columns", None)

        ### Meta data ###
        self.filename = kwargs.pop("filename", None)
        self.sampling_freq = kwargs.pop("sampling_freq", None)
        self.detector = kwargs.pop("detector", None)
        self.features = kwargs.pop("features", None)
        self.sessions = kwargs.pop("sessions", None)

        super().__init__(*args, **kwargs)
        if self.sessions is not None:
            if not len(self.sessions) == len(self):
                raise ValueError("Make sure sessions is same length as data.")
            self.sessions = np.array(self.sessions)
        # if (self.fex_columns is None) and (not self._metadata):
        #     try:
        #         self.fex_columns = self._metadata
        #     except:
        #         print('Failed to import _metadata to fex_columns')

        # Set _metadata attributes on series: Kludgy solution
        for k in self:
            self[k].sampling_freq = self.sampling_freq
            self[k].sessions = self.sessions

    @property
    def _constructor(self):
        return Fex

    @property
    def _constructor_sliced(self):
        return FexSeries

    def _ixs(self, i, axis=0):
        """Override indexing to ensure Fex._metadata is propogated correctly
            when integer indexing

        i : int, slice, or sequence of integers
        axis : int
        """
        result = super()._ixs(i, axis=axis)

        # Override columns
        if axis == 0:
            if isinstance(i, slice):
                return self[i]
            else:
                label = self.index[i]
                if isinstance(label, Index):
                    # a location index by definition
                    result = self.take(i, axis=axis)
                    copy = True
                else:
                    new_values = self._data.fast_xs(i)

                    # if we are a copy, mark as such
                    copy = (
                        isinstance(new_values, np.ndarray) and new_values.base is None
                    )
                    result = self._constructor_sliced(
                        new_values,
                        index=self.columns,
                        name=self.index[i],
                        dtype=new_values.dtype,
                        au_columns=self.au_columns,
                        emotion_columns=self.emotion_columns,
                        facebox_columns=self.facebox_columns,
                        landmark_columns=self.landmark_columns,
                        facepose_columns=self.facepose_columns,
                        gaze_columns=self.gaze_columns,
                        time_columns=self.time_columns,
                        design_columns=self.design_columns,
                        filename=self.filename,
                        sampling_freq=self.sampling_freq,
                        detector=self.detector,
                        features=self.features,
                        sessions=self.sessions,
                    )
                result._set_is_copy(self, copy=copy)
                return result

        else:
            """
            Notes
            -----
            If slice passed, the resulting data will be a view
            """

            label = self.columns[i]
            if isinstance(i, slice):
                # need to return view
                lab_slice = slice(label[0], label[-1])
                return self.loc[:, lab_slice]
            else:
                if isinstance(label, Index):
                    return self._take(i, axis=1, convert=True)

                index_len = len(self.index)

                # if the values returned are not the same length
                # as the index (iow a not found value), iget returns
                # a 0-len ndarray. This is effectively catching
                # a numpy error (as numpy should really raise)
                values = self._data.iget(i)

                if index_len and not len(values):
                    values = np.array([np.nan] * index_len, dtype=object)
                result = self._constructor_sliced(
                    values,
                    index=self.index,
                    name=label,
                    fastpath=True,
                    au_columns=self.au_columns,
                    emotion_columns=self.emotion_columns,
                    facebox_columns=self.facebox_columns,
                    landmark_columns=self.landmark_columns,
                    facepose_columns=self.facepose_columns,
                    gaze_columns=self.gaze_columns,
                    time_columns=self.time_columns,
                    design_columns=self.design_columns,
                    filename=self.filename,
                    sampling_freq=self.sampling_freq,
                    detector=self.detector,
                    features=self.features,
                    sessions=self.sessions,
                )

                # this is a cached value, mark it so
                result._set_as_cached(label, self)
                return result

    def aus(self):
        """Returns the Action Units data

        Returns:
            DataFrame: Action Units data
        """
        return self[self.au_columns]

    def emotions(self):
        """Returns the emotion data

        Returns:
            DataFrame: emotion data
        """
        return self[self.emotion_columns]

    def landmark(self):
        """Returns the landmark data

        Returns:
            DataFrame: landmark data
        """
        return self[self.landmark_columns]

    def input(self):
        """Returns input column as string

        Returns:
            string: path to input image
        """
        return self["input"]

    def landmark_x(self):
        """Returns the x landmarks.

        Returns:
            DataFrame: x landmarks.
        """
        ######## TODO: NATSORT columns before returning #######
        x_cols = [col for col in self.landmark_columns if "x" in col]
        return self[x_cols]

    def landmark_y(self):
        """Returns the y landmarks.

        Returns:
            DataFrame: y landmarks.
        """
        y_cols = [col for col in self.landmark_columns if "y" in col]
        return self[y_cols]

    def facebox(self):
        """Returns the facebox data

        Returns:
            DataFrame: facebox data
        """
        return self[self.facebox_columns]

    def time(self):
        """Returns the time data

        Returns:
            DataFrame: time data
        """
        return self[self.time_columns]

    def design(self):
        """Returns the design data

        Returns:
            DataFrame: time data
        """
        return self[self.design_columns]

    def read_file(self, *args, **kwargs):
        """Loads file into FEX class

        Returns:
            DataFrame: Fex class
        """
        if self.detector == "FACET":
            return self.read_facet(self.filename)
        elif self.detector == "OpenFace":
            return self.read_openface(self.filename)
        elif self.detector == "Affectiva":
            return self.read_affectiva(self.filename)
        elif self.detector == "Feat":
            return self.read_feat(self.filename)
        else:
            print("Must specifiy which detector [Feat, FACET, OpenFace, or Affectiva]")

    def info(self):
        """Print class meta data."""
        attr_list = []
        for name in self._metadata:
            attr_list.append(name + ": " + str(getattr(self, name, None)) + "\n")
        print(f"{self.__class__}\n" + "".join(attr_list))

    ###   Class Methods   ###
    def read_feat(self, filename=None, *args, **kwargs):
        # Check if filename exists in metadata.
        if not filename:
            try:
                filename = self.filename
            except:
                print("filename must be specified.")
        result = read_feat(filename, *args, **kwargs)
        return result

    def read_facet(self, filename=None, *args, **kwargs):
        # Check if filename exists in metadata.
        if not filename:
            try:
                filename = self.filename
            except:
                print("filename must be specified.")
        result = read_facet(filename, *args, **kwargs)
        for name in self._metadata:
            attr_value = getattr(self, name, None)
            if attr_value and getattr(result, name, None) == None:
                setattr(result, name, attr_value)
        return result

    def read_openface(self, filename=None, *args, **kwargs):
        if not filename:
            try:
                filename = self.filename
            except:
                print("filename must be specified.")
        result = read_openface(filename, *args, **kwargs)
        for name in self._metadata:
            attr_value = getattr(self, name, None)
            if attr_value and getattr(result, name, None) == None:
                setattr(result, name, attr_value)
        return result

    def read_affectiva(self, filename=None, *args, **kwargs):
        if not filename:
            try:
                filename = self.filename
            except:
                print("filename must be specified.")
        result = read_affectiva(filename, *args, **kwargs)
        for name in self._metadata:
            attr_value = getattr(self, name, None)
            if attr_value and getattr(result, name, None) == None:
                setattr(result, name, attr_value)
        return result

    def itersessions(self):
        """Iterate over Fex sessions as (session, series) pairs.

        Returns:
            it: a generator that iterates over the sessions of the fex instance

        """
        for x in np.unique(self.sessions):
            yield x, self.loc[self.sessions == x, :]

    def append(self, data, session_id=None, axis=0):
        """Append a new Fex object to an existing object

        Args:
            data: (Fex) Fex instance to append
            session_id: session label
            axis: ([0,1]) Axis to append. Rows=0, Cols=1
        Returns:
            Fex instance
        """
        if not isinstance(data, self.__class__):
            raise ValueError("Make sure data is a Fex instance.")

        if self.empty:
            out = data.copy()
            if session_id is not None:
                out.sessions = np.repeat(session_id, len(data))
        else:
            if self.sampling_freq != data.sampling_freq:
                raise ValueError(
                    "Make sure Fex objects have the same " "sampling frequency"
                )
            if axis == 0:
                out = self.__class__(
                    pd.concat([self, data], axis=axis, ignore_index=True),
                    sampling_freq=self.sampling_freq,
                ).__finalize__(self)
                if session_id is not None:
                    out.sessions = np.hstack(
                        [self.sessions, np.repeat(session_id, len(data))]
                    )
                if self.features is not None:
                    if data.features is not None:
                        if self.features.shape[1] == data.features.shape[1]:
                            out.features = self.features.append(
                                data.features, ignore_index=True
                            )
                        else:
                            raise ValueError(
                                "Different number of features in new dataset."
                            )
                    else:
                        out.features = self.features
                elif data.features is not None:
                    out = data.features
            elif axis == 1:
                out = self.__class__(
                    pd.concat([self, data], axis=axis), sampling_freq=self.sampling_freq
                ).__finalize__(self)
                if self.sessions is not None:
                    if data.sessions is not None:
                        if np.array_equal(self.sessions, data.sessions):
                            out.sessions = self.sessions
                        else:
                            raise ValueError("Both sessions must be identical.")
                    else:
                        out.sessions = self.sessions
                elif data.sessions is not None:
                    out.sessions = data.sessions
                if self.features is not None:
                    out.features = self.features
                    if data.features is not None:
                        out.features.append(data.features, axis=1, ignore_index=True)
                elif data.features is not None:
                    out.features = data.features
            else:
                raise ValueError("Axis must be 1 or 0.")
        return out

    def regress(self, X, y, fit_intercept=True, *args, **kwargs):
        """Regress using nltools.stats.regress.

        fMRI-like regression to predict Fex activity (y) from set of regressors (X).

        Args:
            X (list or str): Independent variable to predict.
            y (list or str): Dependent variable to be predicted.
            fit_intercept (bool): Whether to add intercept before fitting. Defaults to True.

        Returns:
            betas, t-stats, p-values, df, residuals
        """
        if type(X) == list:
            mX = self[X]
        else:
            mX = X

        if fit_intercept:
            mX["intercept"] = 1

        if type(y) == str:
            my = self[y]
        else:
            my = y
        b, t, p, df, res = regress(mX, my, *args, **kwargs)
        b_df = pd.DataFrame(b, index=mX.columns, columns=my.columns)
        t_df = pd.DataFrame(t, index=mX.columns, columns=my.columns)
        p_df = pd.DataFrame(p, index=mX.columns, columns=my.columns)
        df_df = pd.DataFrame([df], index=[0], columns=my.columns)
        res_df = pd.DataFrame(res, columns=my.columns)
        return b_df, t_df, p_df, df_df, res_df

    def ttest(self, popmean=0, threshold_dict=None):
        """Conducts 1 sample ttest.

        Args:
            popmean (int, optional): [description]. Defaults to 0.
            threshold_dict ([type], optional): [description]. Defaults to None.

        Returns:
            [type]: [description]
        """
        return ttest_1samp(self, popmean)

    def predict(self, X, y, model=LinearRegression, *args, **kwargs):
        """[summary]

        Args:
            X (list or DataFrame): List of column names or dataframe to be used as features for prediction
            y (string or array): y values to be predicted
            model (class, optional): Any sklearn model. Defaults to LinearRegression.
            *args, **kwargs: Model arguments

        Returns:
            model: Fit model instance.
        """
        if type(X) == list:
            mX = self[X]
        else:
            mX = X

        if type(y) == str:
            my = self[y]
        else:
            my = y
        clf = model(*args, **kwargs)
        clf.fit(mX, my)
        return clf

    def downsample(self, target, **kwargs):
        """Downsample Fex columns. Relies on nltools.stats.downsample,
           but ensures that returned object is a Fex object.

        Args:
            target(float): downsampling target, typically in samples not
                            seconds
            kwargs: additional inputs to nltools.stats.downsample

        """

        df_ds = downsample(
            self, sampling_freq=self.sampling_freq, target=target, **kwargs
        ).__finalize__(self)
        df_ds.sampling_freq = target

        if self.features is not None:
            ds_features = downsample(
                self.features, sampling_freq=self.sampling_freq, target=target, **kwargs
            )
        else:
            ds_features = self.features
        df_ds.features = ds_features
        return df_ds
        # return self.__class__(df_ds, sampling_freq=target, features=ds_features)

    def upsample(self, target, target_type="hz", **kwargs):
        """Upsample Fex columns. Relies on nltools.stats.upsample,
            but ensures that returned object is a Fex object.

        Args:
            target(float): upsampling target, default 'hz' (also 'samples',
                           'seconds')
            kwargs: additional inputs to nltools.stats.upsample

        """

        df_us = upsample(
            self,
            sampling_freq=self.sampling_freq,
            target=target,
            target_type=target_type,
            **kwargs,
        )
        if self.features is not None:
            us_features = upsample(
                self.features,
                sampling_freq=self.sampling_freq,
                target=target,
                target_type=target_type,
                **kwargs,
            )
        else:
            us_features = self.features
        return self.__class__(df_us, sampling_freq=target, features=us_features)

    def distance(self, method="euclidean", **kwargs):
        """Calculate distance between rows within a Fex() instance.

        Args:
            method: type of distance metric (can use any scikit learn or
                    sciypy metric)

        Returns:
            dist: Outputs a 2D distance matrix.

        """

        return Adjacency(
            pairwise_distances(self, metric=method, **kwargs), matrix_type="Distance"
        )

    def rectification(self, std=3):
        """Removes time points when the face position moved
        more than N standard deviations from the mean.

        Args:
            std (default 3): standard deviation from mean to remove outlier face locations
        Returns:
            data: cleaned FEX object

        """
        #### TODO: CHECK IF FACET OR FIND WAY TO DO WITH OTHER ONES TOO #####
        if self.facebox_columns and self.au_columns and self.emotion_columns:
            cleaned = deepcopy(self)
            face_columns = self.facebox_columns
            x_m = self.FaceRectX.mean()
            x_std = self.FaceRectX.std()
            y_m = self.FaceRectY.mean()
            y_std = self.FaceRectY.std()
            x_bool = (self.FaceRectX > std * x_std + x_m) | (
                self.FaceRectX < x_m - std * x_std
            )
            y_bool = (self.FaceRectY > std * y_std + y_m) | (
                self.FaceRectY < y_m - std * y_std
            )
            xy_bool = x_bool | y_bool
            cleaned.loc[
                xy_bool, face_columns + self.au_columns + self.emotion_columns
            ] = np.nan
            return cleaned
        else:
            raise ValueError("Facebox columns need to be defined.")

    def baseline(self, baseline="median", normalize=None, ignore_sessions=False):
        """Reference a Fex object to a baseline.

        Args:
            method: {'median', 'mean', 'begin', FexSeries instance}. Will subtract baseline
                    from Fex object (e.g., mean, median).  If passing a Fex
                    object, it will treat that as the baseline.
            normalize: (str). Can normalize results of baseline.
                        Values can be [None, 'db','pct']; default None.
            ignore_sessions: (bool) If True, will ignore Fex.sessions
                             information. Otherwise, method will be applied
                             separately to each unique session.

        Returns:
            Fex object
        """

        if self.sessions is None or ignore_sessions:
            out = self.copy()
            if type(baseline) == str:
                if baseline == "median":
                    baseline_values = out.median()
                elif baseline == "mean":
                    baseline_values = out.mean()
                elif baseline == "begin":
                    baseline_values = out.iloc[0, :]
                else:
                    raise ValueError(
                        "%s is not implemented please use {mean, median, Fex}"
                        % baseline
                    )
            elif isinstance(baseline, (Series, FexSeries)):
                baseline_values = baseline
            elif isinstance(baseline, (Fex, DataFrame)):
                raise ValueError("Must pass in a FexSeries not a FexSeries Instance.")

            if normalize == "db":
                out = 10 * np.log10(out - baseline_values) / baseline_values
            if normalize == "pct":
                out = 100 * (out - baseline_values) / baseline_values
            else:
                out = out - baseline_values
        else:
            out = self.__class__(sampling_freq=self.sampling_freq)
            for k, v in self.itersessions():
                if type(baseline) == str:
                    if baseline == "median":
                        baseline_values = v.median()
                    elif baseline == "mean":
                        baseline_values = v.mean()
                    elif baseline == "begin":
                        baseline_values = v.iloc[0, :]
                    else:
                        raise ValueError(
                            "%s is not implemented please use {mean, median, Fex}"
                            % baseline
                        )
                elif isinstance(baseline, (Series, FexSeries)):
                    baseline_values = baseline
                elif isinstance(baseline, (Fex, DataFrame)):
                    raise ValueError(
                        "Must pass in a FexSeries not a FexSeries Instance."
                    )

                if normalize == "db":
                    out = out.append(
                        10 * np.log10(v - baseline_values) / baseline_values,
                        session_id=k,
                    )
                if normalize == "pct":
                    out = out.append(
                        100 * (v - baseline_values) / baseline_values, session_id=k
                    )
                else:
                    out = out.append(v - baseline_values, session_id=k)
        return out.__finalize__(self)
        # return self.__class__(out, sampling_freq=self.sampling_freq, features=self.features, sessions=self.sessions)

    def clean(
        self,
        detrend=True,
        standardize=True,
        confounds=None,
        low_pass=None,
        high_pass=None,
        ensure_finite=False,
        ignore_sessions=False,
        *args,
        **kwargs,
    ):

        """Clean Time Series signal

        This function wraps nilearn functionality and can filter, denoise,
        detrend, etc.

        See http://nilearn.github.io/modules/generated/nilearn.signal.clean.html

        This function can do several things on the input signals, in
        the following order:
            - detrend
            - standardize
            - remove confounds
            - low- and high-pass filter

        If Fex.sessions is not None, sessions will be cleaned separately.

        Args:
            confounds: (numpy.ndarray, str or list of Confounds timeseries)
                        Shape must be (instant number, confound number),
                        or just (instant number,). The number of time
                        instants in signals and confounds must be identical
                        (i.e. signals.shape[0] == confounds.shape[0]). If a
                        string is provided, it is assumed to be the name of
                        a csv file containing signals as columns, with an
                        optional one-line header. If a list is provided,
                        all confounds are removed from the input signal,
                        as if all were in the same array.

            low_pass: (float) low pass cutoff frequencies in Hz.
            high_pass: (float) high pass cutoff frequencies in Hz.
            detrend: (bool) If detrending should be applied on timeseries
                     (before confound removal)
            standardize: (bool) If True, returned signals are set to unit
                         variance.
            ensure_finite: (bool) If True, the non-finite values
                           (NANs and infs) found in the data will be
                           replaced by zeros.
            ignore_sessions: (bool) If True, will ignore Fex.sessions
                             information. Otherwise, method will be applied
                             separately to each unique session.
        Returns:
            cleaned Fex instance

        """
        if self.sessions is not None:
            if ignore_sessions:
                sessions = None
            else:
                sessions = self.sessions
        else:
            sessions = None
        return self.__class__(
            pd.DataFrame(
                clean(
                    self.values,
                    detrend=detrend,
                    standardize=standardize,
                    confounds=confounds,
                    low_pass=low_pass,
                    high_pass=high_pass,
                    ensure_finite=ensure_finite,
                    t_r=1.0 / np.float(self.sampling_freq),
                    sessions=sessions,
                    *args,
                    **kwargs,
                ),
                columns=self.columns,
            ),
            sampling_freq=self.sampling_freq,
            features=self.features,
            sessions=self.sessions,
        )

    def decompose(self, algorithm="pca", axis=1, n_components=None, *args, **kwargs):
        """Decompose Fex instance

        Args:
            algorithm: (str) Algorithm to perform decomposition
                        types=['pca','ica','nnmf','fa']
            axis: dimension to decompose [0,1]
            n_components: (int) number of components. If None then retain
                        as many as possible.

        Returns:
            output: a dictionary of decomposition parameters

        """

        out = {}
        out["decomposition_object"] = set_decomposition_algorithm(
            algorithm=algorithm, n_components=n_components, *args, **kwargs
        )
        com_names = ["c%s" % str(x + 1) for x in range(n_components)]
        if axis == 0:
            out["decomposition_object"].fit(self.T)
            out["components"] = self.__class__(
                pd.DataFrame(
                    out["decomposition_object"].transform(self.T),
                    index=self.columns,
                    columns=com_names,
                ),
                sampling_freq=None,
            )
            out["weights"] = self.__class__(
                pd.DataFrame(
                    out["decomposition_object"].components_.T,
                    index=self.index,
                    columns=com_names,
                ),
                sampling_freq=self.sampling_freq,
                features=self.features,
                sessions=self.sessions,
            )
        if axis == 1:
            out["decomposition_object"].fit(self)
            out["components"] = self.__class__(
                pd.DataFrame(
                    out["decomposition_object"].transform(self), columns=com_names
                ),
                sampling_freq=self.sampling_freq,
                features=self.features,
                sessions=self.sessions,
            )
            out["weights"] = self.__class__(
                pd.DataFrame(
                    out["decomposition_object"].components_,
                    index=com_names,
                    columns=self.columns,
                ),
                sampling_freq=None,
            ).T
        return out

    def extract_mean(self, ignore_sessions=False, *args, **kwargs):
        """Extract mean of each feature

        Args:
            ignore_sessions: (bool) ignore sessions or extract separately
                                    by sessions if available.
        Returns:
            Fex: mean values for each feature

        """
        prefix = "mean_"
        if self.sessions is None or ignore_sessions:
            feats = pd.DataFrame(self.mean()).T
        else:
            feats = pd.DataFrame()
            for k, v in self.itersessions():
                feats = feats.append(pd.Series(v.mean(), name=k))
        feats = self.__class__(feats)
        feats.columns = prefix + feats.columns
        feats = feats.__finalize__(self)
        if ignore_sessions == False:
            feats.sessions = np.unique(self.sessions)
        for attr_name in [
            "au_columns",
            "emotion_columns",
            "facebox_columns",
            "landmark_columns",
            "facepose_columns",
            "gaze_columns",
            "time_columns",
        ]:
            attr_list = feats.__getattr__(attr_name)
            if attr_list:
                new_attr = [prefix + attr for attr in attr_list]
                feats.__setattr__(attr_name, new_attr)
        return feats

    def extract_min(self, ignore_sessions=False, *args, **kwargs):
        """Extract minimum of each feature

        Args:
            ignore_sessions: (bool) ignore sessions or extract separately
                                    by sessions if available.
        Returns:
            Fex: (Fex) minimum values for each feature

        """
        prefix = "min_"
        if self.sessions is None or ignore_sessions:
            feats = pd.DataFrame(self.min()).T
        else:
            feats = pd.DataFrame()
            for k, v in self.itersessions():
                feats = feats.append(pd.Series(v.min(), name=k))
        feats = self.__class__(feats)
        feats.columns = prefix + feats.columns
        feats = feats.__finalize__(self)
        if ignore_sessions == False:
            feats.sessions = np.unique(self.sessions)
        for attr_name in [
            "au_columns",
            "emotion_columns",
            "facebox_columns",
            "landmark_columns",
            "facepose_columns",
            "gaze_columns",
            "time_columns",
        ]:
            attr_list = feats.__getattr__(attr_name)
            if attr_list:
                new_attr = [prefix + attr for attr in attr_list]
                feats.__setattr__(attr_name, new_attr)
        return feats

    def extract_max(self, ignore_sessions=False, *args, **kwargs):
        """Extract maximum of each feature

        Args:
            ignore_sessions: (bool) ignore sessions or extract separately
                                    by sessions if available.
        Returns:
            fex: (Fex) maximum values for each feature

        """
        prefix = "max_"
        if self.sessions is None or ignore_sessions:
            feats = pd.DataFrame(self.max()).T
        else:
            feats = pd.DataFrame()
            for k, v in self.itersessions():
                feats = feats.append(pd.Series(v.max(), name=k))
        feats = self.__class__(feats)
        feats.columns = prefix + feats.columns
        feats = feats.__finalize__(self)
        if ignore_sessions == False:
            feats.sessions = np.unique(self.sessions)
        for attr_name in [
            "au_columns",
            "emotion_columns",
            "facebox_columns",
            "landmark_columns",
            "facepose_columns",
            "gaze_columns",
            "time_columns",
        ]:
            attr_list = feats.__getattr__(attr_name)
            if attr_list:
                new_attr = [prefix + attr for attr in attr_list]
                feats.__setattr__(attr_name, new_attr)
        return feats

    def extract_summary(
        self, mean=True, max=True, min=True, ignore_sessions=False, *args, **kwargs
    ):
        """Extract summary of multiple features

        Args:
            mean: (bool) extract mean of features
            max: (bool) extract max of features
            min: (bool) extract min of features
            ignore_sessions: (bool) ignore sessions or extract separately
                                    by sessions if available.

        Returns:
            fex: (Fex)

        """

        out = self.__class__().__finalize__(self)
        if ignore_sessions == False:
            out.sessions = np.unique(self.sessions)
        if mean:
            new = self.extract_mean(ignore_sessions=ignore_sessions, *args, **kwargs)
            out = out.append(new, axis=1)
            # for attr_name in ['au_columns', 'emotion_columns', 'facebox_columns', 'landmark_columns', 'facepose_columns', 'gaze_columns', 'time_columns']:
            #     if new.__getattr__(attr_name):
            #         new_attr = new.__getattr__(attr_name)
            #         out.__setattr__(attr_name, new_attr)
        if max:
            new = self.extract_max(ignore_sessions=ignore_sessions, *args, **kwargs)
            out = out.append(new, axis=1)
            # for attr_name in ['au_columns', 'emotion_columns', 'facebox_columns', 'landmark_columns', 'facepose_columns', 'gaze_columns', 'time_columns']:
            #     if out.__getattr__(attr_name) and new.__getattr__(attr_name):
            #         new_attr = out.__getattr__(attr_name) + new.__getattr__(attr_name)
            #         out.__setattr__(attr_name, new_attr)
        if min:
            new = self.extract_min(ignore_sessions=ignore_sessions, *args, **kwargs)
            out = out.append(new, axis=1)
        for attr_name in [
            "au_columns",
            "emotion_columns",
            "facebox_columns",
            "landmark_columns",
            "facepose_columns",
            "gaze_columns",
            "time_columns",
        ]:
            if self.__getattr__(attr_name):
                new_attr = []
                if mean:
                    new_attr.extend(
                        ["mean_" + attr for attr in self.__getattr__(attr_name)]
                    )
                if max:
                    new_attr.extend(
                        ["max_" + attr for attr in self.__getattr__(attr_name)]
                    )
                if min:
                    new_attr.extend(
                        ["min_" + attr for attr in self.__getattr__(attr_name)]
                    )
                out.__setattr__(attr_name, new_attr)
        return out

    def extract_wavelet(self, freq, num_cyc=3, mode="complex", ignore_sessions=False):
        """Perform feature extraction by convolving with a complex morlet
        wavelet

        Args:
            freq: (float) frequency to extract
            num_cyc: (float) number of cycles for wavelet
            mode: (str) feature to extract, e.g.,
                        ['complex','filtered','phase','magnitude','power']
            ignore_sessions: (bool) ignore sessions or extract separately
                                    by sessions if available.
        Returns:
            convolved: (Fex instance)
        """
        wav = wavelet(freq, sampling_freq=self.sampling_freq, num_cyc=num_cyc)
        if self.sessions is None or ignore_sessions:
            convolved = self.__class__(
                pd.DataFrame(
                    {x: convolve(y, wav, mode="same") for x, y in self.iteritems()}
                ),
                sampling_freq=self.sampling_freq,
            )
        else:
            convolved = self.__class__(sampling_freq=self.sampling_freq)
            for k, v in self.itersessions():
                session = self.__class__(
                    pd.DataFrame(
                        {x: convolve(y, wav, mode="same") for x, y in v.iteritems()}
                    ),
                    sampling_freq=self.sampling_freq,
                )
                convolved = convolved.append(session, session_id=k)
        if mode == "complex":
            convolved = convolved
        elif mode == "filtered":
            convolved = np.real(convolved)
        elif mode == "phase":
            convolved = np.angle(convolved)
        elif mode == "magnitude":
            convolved = np.abs(convolved)
        elif mode == "power":
            convolved = np.abs(convolved) ** 2
        else:
            raise ValueError(
                "Mode must be ['complex','filtered','phase'," "'magnitude','power']"
            )
        convolved = self.__class__(
            convolved,
            sampling_freq=self.sampling_freq,
            features=self.features,
            sessions=self.sessions,
        )
        convolved.columns = (
            "f" + "%s" % round(freq, 2) + "_" + mode + "_" + self.columns
        )
        return convolved

    def extract_multi_wavelet(
        self, min_freq=0.06, max_freq=0.66, bank=8, *args, **kwargs
    ):
        """Convolve with a bank of morlet wavelets. Wavelets are equally
        spaced from min to max frequency. See extract_wavelet for more
        information and options.

        Args:
            min_freq: (float) minimum frequency to extract
            max_freq: (float) maximum frequency to extract
            bank: (int) size of wavelet bank
            num_cyc: (float) number of cycles for wavelet
            mode: (str) feature to extract, e.g.,
                        ['complex','filtered','phase','magnitude','power']
            ignore_sessions: (bool) ignore sessions or extract separately
                                    by sessions if available.
        Returns:
            convolved: (Fex instance)
        """
        out = []
        for f in np.geomspace(min_freq, max_freq, bank):
            out.append(self.extract_wavelet(f, *args, **kwargs))
        return self.__class__(
            pd.concat(out, axis=1),
            sampling_freq=self.sampling_freq,
            features=self.features,
            sessions=self.sessions,
        )

    def extract_boft(self, min_freq=0.06, max_freq=0.66, bank=8, *args, **kwargs):
        """Extract Bag of Temporal features
        Args:
            min_freq: maximum frequency of temporal filters
            max_freq: minimum frequency of temporal filters
            bank: number of temporal filter banks, filters are on exponential scale

        Returns:
            wavs: list of Morlet wavelets with corresponding freq
            hzs:  list of hzs for each Morlet wavelet

        """
        # First generate the wavelets
        target_hz = self.sampling_freq
        freqs = np.geomspace(min_freq, max_freq, bank)
        wavs, hzs = [], []
        for i, f in enumerate(freqs):
            wav = np.real(wavelet(f, sampling_freq=target_hz))
            wavs.append(wav)
            hzs.append(str(np.round(freqs[i], 2)))
        wavs = np.array(wavs)[::-1]
        hzs = np.array(hzs)[::-1]
        # # check asymptotes at lowest freq
        # asym = wavs[-1,:10].sum()
        # if asym > .001:
        #     print("Lowest frequency asymptotes at %2.8f " %(wavs[-1,:10].sum()))

        # Convolve data with wavelets
        Feats2Use = self.columns
        feats = pd.DataFrame()
        for feat in Feats2Use:
            _d = self[[feat]].T
            assert _d.isnull().sum().any() == 0, "Data contains NaNs. Cannot convolve. "
            for iw, cm in enumerate(wavs):
                convolved = np.apply_along_axis(
                    lambda m: np.convolve(m, cm, mode="full"), axis=1, arr=_d.values
                )
                # Extract bin features.
                out = pd.DataFrame(convolved.T).apply(calc_hist_auc, args=(None))
                # 6 bins hardcoded from calc_hist_auc
                colnames = [
                    "pos" + str(i) + "_hz_" + hzs[iw] + "_" + feat for i in range(6)
                ]
                colnames.extend(
                    ["neg" + str(i) + "_hz_" + hzs[iw] + "_" + feat for i in range(6)]
                )
                out = out.T
                out.columns = colnames
                feats = pd.concat([feats, out], axis=1)
        return self.__class__(
            feats, sampling_freq=self.sampling_freq, features=self.features
        )

    def calc_pspi(self):
        if self.detector == "FACET":
            pspi_aus = ["AU4", "AU6", "AU7", "AU9", "AU10", "AU43"]
            out = (
                self["AU4"]
                + self[["AU6", "AU7"]].max(axis=1)
                + self[["AU9", "AU10"]].max(axis=1)
                + self["AU43"]
            )
        if self.detector == "OpenFace":
            out = (
                self["AU04_r"]
                + self[["AU06_r", "AU07_r"]].max(axis=1)
                + self[["AU09_r", "AU10_r"]].max(axis=1)
                + self["AU45_r"]
            )
        return out.__finalize__(self)

    def plot_aus(
        self,
        row_n,
        model=None,
        vectorfield=None,
        muscles=None,
        ax=None,
        color="k",
        linewidth=1,
        linestyle="-",
        gaze=None,
        *args,
        **kwargs,
    ):
        if self.detector == "FACET":
            feats = [
                "AU1",
                "AU2",
                "AU4",
                "AU5",
                "AU6",
                "AU7",
                "AU9",
                "AU10",
                "AU12",
                "AU14",
                "AU15",
                "AU17",
                "AU18",
                "AU20",
                "AU23",
                "AU24",
                "AU25",
                "AU26",
                "AU28",
                "AU43",
                "Pitch",
                "Roll",
                "Yaw",
            ]
            if row_n > len(self):
                raise ValueError("Row number out of range.")
            try:
                au = []
                for feat in feats:
                    aun = self[feat]
                    au.append(aun.copy()[row_n])
                au = np.array(au)
                if model is None:
                    model = load_h5("facet.h5")
                if muscles is not None:
                    muscles["facet"] = 1
                ax = plot_face(
                    model=model,
                    au=au,
                    vectorfield=vectorfield,
                    muscles=muscles,
                    ax=ax,
                    color=color,
                    linewidth=linewidth,
                    linestyle=linestyle,
                    gaze=gaze,
                    *args,
                    **kwargs,
                )
                return ax
            except Exception as e:
                print("Unable to plot data:", e)
        if self.detector == "OpenFace":
            feats = [
                "AU01_r",
                "AU02_r",
                "AU04_r",
                "AU05_r",
                "AU06_r",
                "AU07_r",
                "AU09_r",
                "AU10_r",
                "AU12_r",
                "AU14_r",
                "AU15_r",
                "AU17_r",
                "AU20_r",
                "AU23_r",
                "AU25_r",
                "AU26_r",
                "AU45_r",
            ]
            if row_n > len(self):
                raise ValueError("Row number out of range.")
            try:
                au = []
                for feat in feats:
                    aun = self[feat]
                    au.append(aun.copy()[row_n])
                au = np.array(au + [0, 0, 0])

                if gaze:
                    gaze_dat = ["gaze_0_x", "gaze_0_y", "gaze_1_x", "gaze_1_y"]
                    gaze = []
                    for i in range(4):
                        gaze.append(self[gaze_dat[i]][row_n])
                    if gaze_vecs:
                        gaze.append(1)
                    else:
                        gaze.append(0)
                else:
                    gaze = None

                ax = plot_face(
                    model=model,
                    au=au,
                    vectorfield=vectorfield,
                    muscles=muscles,
                    ax=ax,
                    color=color,
                    linewidth=linewidth,
                    linestyle=linestyle,
                    gaze=gaze,
                    *args,
                    **kwargs,
                )
                return ax
            except Exception as e:
                print("Unable to plot data:", e)
        if self.detector == "Affectiva":
            if "AU01" not in self.au_columns:
                feats = [
                    "innerBrowRaise",
                    "browRaise",
                    "browFurrow",
                    "eyeWiden",
                    "cheekRaise",
                    "lidTighten",
                    "noseWrinkle",
                    "upperLipRaise",
                    "smile",
                    "dimpler",
                    "lipCornerDepressor",
                    "chinRaise",
                    "lipStretch",
                    "lipPress",
                    "mouthOpen",
                    "jawDrop",
                    "eyeClosure",
                ]
            else:
                feats = [
                    "AU01",
                    "AU02",
                    "AU04",
                    "AU05",
                    "AU06",
                    "AU07",
                    "AU09",
                    "AU10",
                    "AU12",
                    "AU14",
                    "AU15",
                    "AU17",
                    "AU20",
                    "AU24",
                    "AU25",
                    "AU26",
                    "AU43",
                ]
            if row_n > len(self):
                raise ValueError("Row number out of range.")
            try:
                au = []
                for feat in feats:
                    aun = self[feat]
                    au.append(aun.copy()[row_n] / 20)
                au = np.array(au + [0, 0, 0])
                ax = plot_face(
                    model=model,
                    au=au,
                    vectorfield=vectorfield,
                    muscles=muscles,
                    ax=ax,
                    color=color,
                    linewidth=linewidth,
                    linestyle=linestyle,
                    gaze=gaze,
                    *args,
                    **kwargs,
                )
                return ax
            except Exception as e:
                print("Unable to plot data:", e)

    def plot_detections(self, draw_landmarks=True, draw_facelines=True, muscle=False):
        """Plots detection results by Feat.

        Args:

        Returns:
            ax
        """
        from PIL import Image
        import matplotlib.pyplot as plt
        from matplotlib.patches import Rectangle

        # check how many images.
        inputs = self.input().unique()
        all_axes = []
        for imagefile in inputs:
            f, axes = plt.subplots(1, 3, figsize=(15, 7))
            ax = axes[0]
            try:
                if os.path.exists(imagefile):
                    color = "w"
                    # draw base image
                    im = Image.open(imagefile)
                    ax.imshow(im)
                    image_exists = True
                else:
                    image_exists = False
                    color = "k"
            except:
                color = "k"
                print(f"Input image {imagefile} not found.")
                image_exists = False

            sub_data = self.query("input==@imagefile")
            for i in range(len(sub_data)):
                # draw landmarks
                row = sub_data.iloc[[i]]
                landmark = row.landmark().values[0]
                currx = landmark[:68]
                curry = landmark[68:]
                if draw_landmarks:
                    if draw_facelines:
                        draw_lineface(currx, curry, ax=ax, color=color, linewidth=3)
                    else:
                        draw_lineface(currx, curry, ax=ax, color=color, linewidth=0)
                # muscle
                if muscle:
                    au20index = [
                        f"AU{str(i).zfill(2)}"
                        for i in [
                            1,
                            2,
                            4,
                            5,
                            6,
                            7,
                            9,
                            10,
                            12,
                            14,
                            15,
                            17,
                            18,
                            20,
                            23,
                            24,
                            25,
                            26,
                            28,
                            43,
                        ]
                    ]
                    aus = row.aus().T.reindex(index=au20index).fillna(0).T.values[0]
                    draw_muscles(currx, curry, au=aus, ax=ax, all="heatmap")
                # facebox
                facebox = row.facebox().values[0]
                rect = Rectangle(
                    (facebox[0], facebox[1]),
                    facebox[2],
                    facebox[3],
                    linewidth=2,
                    edgecolor="cyan",
                    fill=False,
                )
                ax.add_patch(rect)

            if image_exists:
                ax.set_title(
                    sub_data.input().unique()[0], loc="center", wrap=True, fontsize=10
                )
            else:
                ax.set(title=imagefile, ylim=ax.get_ylim()[::-1])
                ax.set_aspect("equal", "box")

            # plot AUs
            sub_data.aus().T.plot(kind="barh", ax=axes[1])
            axes[1].invert_yaxis()
            axes[1].get_legend().remove()
            axes[1].set(xlim=[0, 1.1], title="Action Units")

            # plot emotions
            sub_data.emotions().T.plot(kind="barh", ax=axes[2])
            axes[2].invert_yaxis()
            axes[2].get_legend().remove()
            axes[2].set(xlim=[0, 1.1], title="Emotions")

            plt.tight_layout()
            plt.show()
            all_axes.append(axes)
        return axes


class Fextractor:

    """
    Fextractor is a class that extracts and merges features from a Fex instance
    in preparation for data analysis.
    """

    def __init__(self):
        self.extracted_features = []

    def mean(self, fex_object, ignore_sessions=False, *args, **kwargs):
        """Extract mean of each feature

        Args:
            fex_object: (Fex) Fex instance to extract features from.
            ignore_sessions: (bool) ignore sessions or extract separately
                                    by sessions if available.
        Returns:
            Fex: mean values for each feature

        """

        if not isinstance(fex_object, (Fex, DataFrame)):
            raise ValueError("Must pass in a Fex object.")
        self.extracted_features.append(
            fex_object.extract_mean(ignore_sessions, *args, **kwargs)
        )

    def max(self, fex_object, ignore_sessions=False, *args, **kwargs):
        """Extract maximum of each feature

        Args:
            fex_object: (Fex) Fex instance to extract features from.
            ignore_sessions: (bool) ignore sessions or extract separately
                             by sessions if available.
        Returns:
            Fex: (Fex) maximum values for each feature
        """

        if not isinstance(fex_object, (Fex, DataFrame)):
            raise ValueError("Must pass in a Fex object.")
        self.extracted_features.append(
            fex_object.extract_max(ignore_sessions, *args, **kwargs)
        )

    def min(self, fex_object, ignore_sessions=False, *args, **kwargs):
        """Extract minimum of each feature

        Args:
            fex_object: (Fex) Fex instance to extract features from.
            ignore_sessions: (bool) ignore sessions or extract separately
                                    by sessions if available.
        Returns:
            Fex: (Fex) minimum values for each feature

        """

        if not isinstance(fex_object, (Fex, DataFrame)):
            raise ValueError("Must pass in a Fex object.")
        self.extracted_features.append(
            fex_object.extract_min(ignore_sessions, *args, **kwargs)
        )

    def summary(
        self,
        fex_object,
        mean=False,
        max=False,
        min=False,
        ignore_sessions=False,
        *args,
        **kwargs,
    ):
        """Extract summary of multiple features
        Args:
            fex_object: (Fex) Fex instance to extract features from.
            mean: (bool) extract mean of features
            max: (bool) extract max of features
            min: (bool) extract min of features
            ignore_sessions: (bool) ignore sessions or extract separately
                                    by sessions if available.
        Returns:
            fex: (Fex)
        """

        self.extracted_features.append(
            fex_object.extract_summary(mean, max, min, ignore_sessions, *args, **kwargs)
        )

    def wavelet(
        self, fex_object, freq, num_cyc=3, mode="complex", ignore_sessions=False
    ):
        """Perform feature extraction by convolving with a complex morlet
        wavelet

        Args:
            fex_object: (Fex) Fex instance to extract features from.
            freq: (float) frequency to extract
            num_cyc: (float) number of cycles for wavelet
            mode: (str) feature to extract, e.g.,
                        ['complex','filtered','phase','magnitude','power']
            ignore_sessions: (bool) ignore sessions or extract separately
                                    by sessions if available.
        Returns:
            convolved: (Fex instance)
        """

        if not isinstance(fex_object, (Fex, DataFrame)):
            raise ValueError("Must pass in a Fex object.")
        self.extracted_features.append(
            fex_object.extract_wavelet(freq, num_cyc, mode, ignore_sessions)
        )

    def multi_wavelet(
        self, fex_object, min_freq=0.06, max_freq=0.66, bank=8, *args, **kwargs
    ):
        """Convolve with a bank of morlet wavelets. Wavelets are equally
        spaced from min to max frequency. See extract_wavelet for more
        information and options.

        Args:
            fex_object: (Fex) Fex instance to extract features from.
            min_freq: (float) minimum frequency to extract
            max_freq: (float) maximum frequency to extract
            bank: (int) size of wavelet bank
            num_cyc: (float) number of cycles for wavelet
            mode: (str) feature to extract, e.g.,
                        ['complex','filtered','phase','magnitude','power']
            ignore_sessions: (bool) ignore sessions or extract separately
                                    by sessions if available.
        Returns:
            convolved: (Fex instance)
        """

        if not isinstance(fex_object, (Fex, DataFrame)):
            raise ValueError("Must pass in a Fex object.")
        self.extracted_features.append(
            fex_object.extract_multi_wavelet(min_freq, max_freq, bank, *args, **kwargs)
        )

    def boft(self, fex_object, min_freq=0.06, max_freq=0.66, bank=8, *args, **kwargs):
        """Extract Bag of Temporal features
        Args:
            fex_object: (Fex) Fex instance to extract features from.
            min_freq: maximum frequency of temporal filters
            max_freq: minimum frequency of temporal filters
            bank: number of temporal filter banks, filters are on exponential scale

        Returns:
            wavs: list of Morlet wavelets with corresponding freq
            hzs:  list of hzs for each Morlet wavelet

        """

        if not isinstance(fex_object, (Fex, DataFrame)):
            raise ValueError("Must pass in a Fex object.")
        self.extracted_features.append(
            fex_object.extract_boft(min_freq, max_freq, bank, *args, **kwargs)
        )

    def merge(self, out_format="long"):
        """Merge all extracted features to a single dataframe
        Args:
            format: (str) Output format of merged data. Can be set to 'long' or 'wide'. Defaults to long.
        Returns:
            merged: (DataFrame) DataFrame containing merged features extracted
                    from a Fex instance.

        """
        out = reduce(
            lambda x, y: pd.merge(x, y, left_index=True, right_index=True),
            self.extracted_features,
        )
        out["sessions"] = out.index

        if out_format == "long":
            out = out.melt(id_vars="sessions")
        elif out_format == "wide":
            pass
        return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `feat` package."""

import pytest
import pandas as pd
from pandas import DataFrame, Series
import numpy as np
from os.path import join, exists
from .utils import get_test_data_path
from feat.utils import (
    read_facet,
    read_openface,
    read_affectiva,
    registration,
    neutral,
    softmax,
    load_h5,
    load_pickled_model,
)
from nltools.data import Adjacency
import unittest


def test_utils():
    sample = read_openface(join(get_test_data_path(), "OpenFace_Test.csv"))
    lm_cols = ["x_" + str(i) for i in range(0, 68)] + [
        "y_" + str(i) for i in range(0, 68)
    ]
    sample_face = np.array([sample[lm_cols].values[0]])
    registered_lm = registration(sample_face)
    assert registered_lm.shape == (1, 136)

    with pytest.raises(ValueError):
        registration(sample_face, method="badmethod")
    with pytest.raises(TypeError):
        registration(sample_face, method=np.array([1, 2, 3, 4]))
    with pytest.raises(AssertionError):
        registration([sample_face[0]])
    with pytest.raises(AssertionError):
        registration(sample_face[0])
    with pytest.raises(AssertionError):
        registration(sample_face[:, :-1])

    # Test softmax
    assert softmax(0) == 0.5
    # Test badfile.
    with pytest.raises(Exception):
        load_h5("badfile.h5")

    # Test loading of pickled model
    out = load_pickled_model()
    with pytest.raises(Exception):
        load_pickled_model("badfile.pkl")



from __future__ import division

"""Plotting Functions"""

import numpy as np
from sklearn.cross_decomposition import PLSRegression
import matplotlib.pyplot as plt
from feat.utils import load_h5

# from utils import load_h5
import warnings
import seaborn as sns
import matplotlib.colors as colors
from collections import OrderedDict

__all__ = [
    "draw_lineface",
    "plot_face",
    "draw_vectorfield",
    "draw_muscles",
    "get_heat",
    "predict",
]
__author__ = ["Sophie Byrne", "Jin Hyun Cheong", "Luke Chang"]


def draw_lineface(
    currx,
    curry,
    ax=None,
    color="k",
    linestyle="-",
    linewidth=1,
    gaze=None,
    *args,
    **kwargs
):
    """Plot Line Face

    Args:
        currx: vector (len(68)) of x coordinates
        curry: vector (len(68)) of y coordinates
        ax: matplotlib axis to add
        color: matplotlib line color
        linestyle: matplotlib linestyle
        linewidth: matplotlib linewidth
        gaze: array (len(4)) of gaze vectors (fifth value is whether to draw vectors)
    """

    face_outline = plt.Line2D(
        [
            currx[0],
            currx[1],
            currx[2],
            currx[3],
            currx[4],
            currx[5],
            currx[6],
            currx[7],
            currx[8],
            currx[9],
            currx[10],
            currx[11],
            currx[12],
            currx[13],
            currx[14],
            currx[15],
            currx[16],
        ],
        [
            curry[0],
            curry[1],
            curry[2],
            curry[3],
            curry[4],
            curry[5],
            curry[6],
            curry[7],
            curry[8],
            curry[9],
            curry[10],
            curry[11],
            curry[12],
            curry[13],
            curry[14],
            curry[15],
            curry[16],
        ],
        color=color,
        linestyle=linestyle,
        linewidth=linewidth,
        *args,
        **kwargs
    )

    eye_l = plt.Line2D(
        [currx[36], currx[37], currx[38], currx[39], currx[40], currx[41], currx[36]],
        [curry[36], curry[37], curry[38], curry[39], curry[40], curry[41], curry[36]],
        color=color,
        linestyle=linestyle,
        linewidth=linewidth,
        *args,
        **kwargs
    )

    eye_r = plt.Line2D(
        [currx[42], currx[43], currx[44], currx[45], currx[46], currx[47], currx[42]],
        [curry[42], curry[43], curry[44], curry[45], curry[46], curry[47], curry[42]],
        color=color,
        linestyle=linestyle,
        linewidth=linewidth,
        *args,
        **kwargs
    )

    eyebrow_l = plt.Line2D(
        [currx[17], currx[18], currx[19], currx[20], currx[21]],
        [curry[17], curry[18], curry[19], curry[20], curry[21]],
        color=color,
        linestyle=linestyle,
        linewidth=linewidth,
        *args,
        **kwargs
    )

    eyebrow_r = plt.Line2D(
        [currx[22], currx[23], currx[24], currx[25], currx[26]],
        [curry[22], curry[23], curry[24], curry[25], curry[26]],
        color=color,
        linestyle=linestyle,
        linewidth=linewidth,
        *args,
        **kwargs
    )

    lips1 = plt.Line2D(
        [
            currx[48],
            currx[49],
            currx[50],
            currx[51],
            currx[52],
            currx[53],
            currx[54],
            currx[64],
            currx[63],
            currx[62],
            currx[61],
            currx[60],
            currx[48],
        ],
        [
            curry[48],
            curry[49],
            curry[50],
            curry[51],
            curry[52],
            curry[53],
            curry[54],
            curry[64],
            curry[63],
            curry[62],
            curry[61],
            curry[60],
            curry[48],
        ],
        color=color,
        linestyle=linestyle,
        linewidth=linewidth,
        *args,
        **kwargs
    )

    lips2 = plt.Line2D(
        [
            currx[48],
            currx[60],
            currx[67],
            currx[66],
            currx[65],
            currx[64],
            currx[54],
            currx[55],
            currx[56],
            currx[57],
            currx[58],
            currx[59],
            currx[48],
        ],
        [
            curry[48],
            curry[60],
            curry[67],
            curry[66],
            curry[65],
            curry[64],
            curry[54],
            curry[55],
            curry[56],
            curry[57],
            curry[58],
            curry[59],
            curry[48],
        ],
        color=color,
        linestyle=linestyle,
        linewidth=linewidth,
        *args,
        **kwargs
    )

    nose1 = plt.Line2D(
        [currx[27], currx[28], currx[29], currx[30]],
        [curry[27], curry[28], curry[29], curry[30]],
        color=color,
        linestyle=linestyle,
        linewidth=linewidth,
        *args,
        **kwargs
    )

    nose2 = plt.Line2D(
        [currx[31], currx[32], currx[33], currx[34], currx[35]],
        [curry[31], curry[32], curry[33], curry[34], curry[35]],
        color=color,
        linestyle=linestyle,
        linewidth=linewidth,
        *args,
        **kwargs
    )
    if gaze is None:
        gaze = [0, 0, 0, 0]

    else:
        if len(gaze) != 4:
            raise ValueError("gaze must be len(4).")
        gaze = [gaze[0], gaze[1] / 2, gaze[2], gaze[3] / 2]  # , gaze[4]]

    x = (currx[37] + currx[38] + currx[41] + currx[40]) / 4
    y = (curry[37] + curry[38] + curry[40] + curry[41]) / 4
    width = (-curry[37] - curry[38] + curry[40] + curry[41]) / 5
    pupil_l = plt.Circle([x + gaze[0], y - gaze[1]], width, color="k")
    x1 = (currx[43] + currx[46] + currx[44] + currx[47]) / 4
    y1 = (curry[43] + curry[44] + curry[46] + curry[47]) / 4
    width = (-curry[43] - curry[44] + curry[46] + curry[47]) / 5
    pupil_r = plt.Circle([x1 + gaze[2], y1 - gaze[3]], width, color="k")

    if ax is None:
        ax = _create_empty_figure()

    ax.add_patch(pupil_l)
    ax.add_patch(pupil_r)
    ax.add_line(face_outline)
    ax.add_line(eye_l)
    ax.add_line(eye_r)
    ax.add_line(eyebrow_l)
    ax.add_line(eyebrow_r)
    ax.add_line(lips1)
    ax.add_line(lips2)
    ax.add_line(nose1)
    ax.add_line(nose2)
    if gaze:
        ax.quiver(
            [x, x1],
            [y, y1],
            [10 * gaze[0], 10 * gaze[2]],
            [-10 * gaze[1], -10 * gaze[3]],
            color="r",
            width=0.005,
            angles="xy",
            scale_units="xy",
            scale=1,
        )


def draw_vectorfield(
    reference, target, color="r", scale=1, width=0.007, ax=None, *args, **kwargs
):
    """Draw vectorfield from reference to target

    Args:
        reference: reference landmarks (2,68)
        target: target landmarks (2,68)
        ax: matplotlib axis instance
        au: vector of action units (len(17))

    """
    if reference.shape != (2, 68):
        raise ValueError("shape of reference must be (2,68)")
    if target.shape != (2, 68):
        raise ValueError("shape of target must be (2,68)")

    currx = []
    curry = []
    for i in range(68):
        currx.append(target[0, i] - reference[0, i])
        curry.append(target[1, i] - reference[1, i])

        if ax is None:
            ax = _create_empty_figure()

    ax.quiver(
        reference[0, :],
        reference[1, :],
        currx,
        curry,
        color=color,
        width=width,
        angles="xy",
        scale_units="xy",
        scale=scale,
        *args,
        **kwargs
    )


def draw_muscles(currx, curry, au=None, ax=None, *args, **kwargs):
    """Draw Muscles

    Args:
        currx: vector (len(68)) of x coordinates
        curry: vector (len(68)) of y coordinates
        ax: matplotlib axis to add
    """
    masseter_l = plt.Polygon(
        [
            [currx[2], curry[2]],
            [currx[3], curry[3]],
            [currx[4], curry[4]],
            [currx[5], curry[5]],
            [currx[6], curry[6]],
            [currx[5], curry[33]],
        ]
    )

    masseter_r = plt.Polygon(
        [
            [currx[14], curry[14]],
            [currx[13], curry[13]],
            [currx[12], curry[12]],
            [currx[11], curry[11]],
            [currx[10], curry[10]],
            [currx[11], curry[33]],
        ]
    )

    temporalis_l = plt.Polygon(
        [
            [currx[2], curry[2]],
            [currx[1], curry[1]],
            [currx[0], curry[0]],
            [currx[17], curry[17]],
            [currx[36], curry[36]],
        ]
    )

    temporalis_r = plt.Polygon(
        [
            [currx[14], curry[14]],
            [currx[15], curry[15]],
            [currx[16], curry[16]],
            [currx[26], curry[26]],
            [currx[45], curry[45]],
        ]
    )

    dep_lab_inf_l = plt.Polygon(
        [
            [currx[57], curry[57]],
            [currx[58], curry[58]],
            [currx[59], curry[59]],
            [currx[6], curry[6]],
            [currx[7], curry[7]],
        ],
        fill=True,
    )

    dep_lab_inf_r = plt.Polygon(
        [
            [currx[57], curry[57]],
            [currx[56], curry[56]],
            [currx[55], curry[55]],
            [currx[10], curry[10]],
            [currx[9], curry[9]],
        ],
        fill=True,
    )

    dep_ang_or_r = plt.Polygon(
        [[currx[54], curry[54]], [currx[9], curry[9]], [currx[10], curry[10]]],
        fill=True,
    )

    dep_ang_or_l = plt.Polygon(
        [[currx[48], curry[48]], [currx[7], curry[7]], [currx[6], curry[6]]], fill=True
    )

    mentalis_l = plt.Polygon(
        [[currx[58], curry[58]], [currx[7], curry[7]], [currx[8], curry[8]]], fill=True
    )

    mentalis_r = plt.Polygon(
        [[currx[56], curry[56]], [currx[9], curry[9]], [currx[8], curry[8]]], fill=True
    )

    risorius_l = plt.Polygon(
        [[currx[4], curry[4]], [currx[5], curry[5]], [currx[48], curry[48]]], fill=True
    )

    risorius_r = plt.Polygon(
        [[currx[11], curry[11]], [currx[12], curry[12]], [currx[54], curry[54]]],
        fill=True,
    )

    bottom = (curry[8] - curry[57]) / 2
    orb_oris_l = plt.Polygon(
        [
            [currx[48], curry[48]],
            [currx[59], curry[59]],
            [currx[58], curry[58]],
            [currx[57], curry[57]],
            [currx[56], curry[56]],
            [currx[55], curry[55]],
            [currx[54], curry[54]],
            [currx[55], curry[55] + bottom],
            [currx[56], curry[56] + bottom],
            [currx[57], curry[57] + bottom],
            [currx[58], curry[58] + bottom],
            [currx[59], curry[59] + bottom],
        ]
    )

    orb_oris_u = plt.Polygon(
        [
            [currx[48], curry[48]],
            [currx[49], curry[49]],
            [currx[50], curry[50]],
            [currx[51], curry[51]],
            [currx[52], curry[52]],
            [currx[53], curry[53]],
            [currx[54], curry[54]],
            [currx[33], curry[33]],
        ],
        fill=True,
    )

    frontalis_l = plt.Polygon(
        [
            [currx[27], curry[27]],
            [currx[39], curry[39]],
            [currx[38], curry[38]],
            [currx[37], curry[37]],
            [currx[36], curry[36]],
            [currx[17], curry[17]],
            [currx[18], curry[18]],
            [currx[19], curry[19]],
            [currx[20], curry[20]],
            [currx[21], curry[21]],
        ]
    )

    frontalis_r = plt.Polygon(
        [
            [currx[27], curry[27]],
            [currx[22], curry[22]],
            [currx[23], curry[23]],
            [currx[24], curry[24]],
            [currx[25], curry[25]],
            [currx[26], curry[26]],
            [currx[45], curry[45]],
            [currx[44], curry[44]],
            [currx[43], curry[43]],
            [currx[42], curry[42]],
        ]
    )

    frontalis_inner_l = plt.Polygon(
        [[currx[27], curry[27]], [currx[39], curry[39]], [currx[21], curry[21]]]
    )

    frontalis_inner_r = plt.Polygon(
        [[currx[27], curry[27]], [currx[42], curry[42]], [currx[22], curry[22]]]
    )

    cor_sup_l = plt.Polygon(
        [[currx[28], curry[28]], [currx[19], curry[19]], [currx[20], curry[20]]]
    )

    cor_sup_r = plt.Polygon(
        [[currx[28], curry[28]], [currx[23], curry[23]], [currx[24], curry[24]]]
    )

    lev_lab_sup_l = plt.Polygon(
        [[currx[41], curry[41]], [currx[40], curry[40]], [currx[49], curry[49]]]
    )

    lev_lab_sup_r = plt.Polygon(
        [[currx[47], curry[47]], [currx[46], curry[46]], [currx[53], curry[53]]]
    )

    lev_lab_sup_an_l = plt.Polygon(
        [[currx[39], curry[39]], [currx[49], curry[49]], [currx[31], curry[31]]]
    )

    lev_lab_sup_an_r = plt.Polygon(
        [[currx[35], curry[35]], [currx[42], curry[42]], [currx[53], curry[53]]]
    )

    zyg_maj_l = plt.Polygon(
        [[currx[48], curry[48]], [currx[3], curry[3]], [currx[2], curry[2]]], color="r"
    )

    zyg_maj_r = plt.Polygon(
        [[currx[54], curry[54]], [currx[13], curry[13]], [currx[14], curry[14]]],
        color="r",
    )

    width = (curry[21] - curry[39]) / 2
    orb_oc_l = plt.Polygon(
        [
            [currx[36] - width / 3, curry[36] + width / 2],
            [currx[36], curry[36] + width],
            [currx[37], curry[37] + width],
            [currx[38], curry[38] + width],
            [currx[39], curry[39] + width],
            [currx[39] + width / 3, curry[39] + width / 2],
            [currx[39] + width / 2, curry[39]],
            [currx[39] + width / 3, curry[39] - width / 2],
            [currx[39], curry[39] - width],
            [currx[40], curry[40] - width],
            [currx[41], curry[41] - width],
            [currx[36], curry[36] - width],
            [currx[36] - width / 3, curry[36] - width / 2],
            [currx[36] - width / 2, curry[36]],
        ]
    )

    orb_oc_l_inner = plt.Polygon(
        [
            [currx[36] - width / 6, curry[36] + width / 5],
            [currx[36], curry[36] + width / 2],
            [currx[37], curry[37] + width / 2],
            [currx[38], curry[38] + width / 2],
            [currx[39], curry[39] + width / 2],
            [currx[39] + width / 6, curry[39] + width / 5],
            [currx[39] + width / 5, curry[39]],
            [currx[39] + width / 6, curry[39] - width / 5],
            [currx[39], curry[39] - width / 2],
            [currx[40], curry[40] - width / 2],
            [currx[41], curry[41] - width / 2],
            [currx[36], curry[36] - width / 2],
            [currx[36] - width / 6, curry[36] - width / 5],
            [currx[36] - width / 5, curry[36]],
        ],
        color="r",
    )

    width2 = (curry[38] - curry[2]) / 1.5
    orb_oc_l_outer = plt.Polygon(
        [
            [currx[39] + width / 2, curry[39]],
            [currx[39], curry[39] - width],
            [currx[40], curry[40] - width2],
            [currx[41], curry[41] - width2],
            [currx[36], curry[36] - width2],
            [currx[36] - width2 / 3, curry[36] - width2 / 2],
            [currx[36] - width / 2, curry[36]],
        ]
    )

    width = (curry[23] - curry[43]) / 2
    orb_oc_r = plt.Polygon(
        [
            [currx[42] - width / 3, curry[42] + width / 2],
            [currx[42], curry[42] + width],
            [currx[43], curry[43] + width],
            [currx[44], curry[44] + width],
            [currx[45], curry[45] + width],
            [currx[45] + width / 3, curry[45] + width / 2],
            [currx[45] + width / 2, curry[45]],
            [currx[45] + width / 3, curry[45] - width / 2],
            [currx[45], curry[45] - width],
            [currx[46], curry[46] - width],
            [currx[47], curry[47] - width],
            [currx[42], curry[42] - width],
            [currx[42] - width / 3, curry[42] - width / 2],
            [currx[42] - width / 2, curry[42]],
        ]
    )

    orb_oc_r_inner = plt.Polygon(
        [
            [currx[42] - width / 6, curry[42] + width / 5],
            [currx[42], curry[42] + width / 2],
            [currx[43], curry[43] + width / 2],
            [currx[44], curry[44] + width / 2],
            [currx[45], curry[45] + width / 2],
            [currx[45] + width / 6, curry[45] + width / 5],
            [currx[45] + width / 5, curry[45]],
            [currx[45] + width / 6, curry[45] - width / 5],
            [currx[45], curry[45] - width / 2],
            [currx[46], curry[46] - width / 2],
            [currx[47], curry[47] - width / 2],
            [currx[42], curry[42] - width / 2],
            [currx[42] - width / 6, curry[42] - width / 5],
            [currx[42] - width / 5, curry[42]],
        ]
    )

    width2 = (curry[44] - curry[14]) / 1.5
    orb_oc_r_outer = plt.Polygon(
        [
            [currx[42] - width / 2, curry[42]],
            [currx[47], curry[47] - width2],
            [currx[46], curry[46] - width2],
            [currx[45], curry[45] - width2],
            [currx[45] + width2 / 3, curry[45] - width2 / 2],
            [currx[45] + width / 2, curry[45]],
        ]
    )
    bucc_l = plt.Polygon(
        [[currx[48], curry[48]], [currx[5], curry[50]], [currx[5], curry[57]]],
        color="r",
    )
    bucc_r = plt.Polygon(
        [[currx[54], curry[54]], [currx[11], curry[52]], [currx[11], curry[57]]],
        color="r",
    )
    muscles = {
        "bucc_l": bucc_l,
        "bucc_r": bucc_r,
        "masseter_l": masseter_l,
        "masseter_r": masseter_r,
        "temporalis_l": temporalis_l,
        "temporalis_r": temporalis_r,
        "dep_lab_inf_l": dep_lab_inf_l,
        "dep_lab_inf_r": dep_lab_inf_r,
        "dep_ang_or_l": dep_ang_or_l,
        "dep_ang_or_r": dep_ang_or_r,
        "mentalis_l": mentalis_l,
        "mentalis_r": mentalis_r,
        "risorius_l": risorius_l,
        "risorius_r": risorius_r,
        "frontalis_l": frontalis_l,
        "frontalis_inner_l": frontalis_inner_l,
        "frontalis_r": frontalis_r,
        "frontalis_inner_r": frontalis_inner_r,
        "cor_sup_r": cor_sup_r,
        "orb_oc_l_outer": orb_oc_l_outer,
        "orb_oc_r_outer": orb_oc_r_outer,
        "lev_lab_sup_l": lev_lab_sup_l,
        "lev_lab_sup_r": lev_lab_sup_r,
        "lev_lab_sup_an_l": lev_lab_sup_an_l,
        "lev_lab_sup_an_r": lev_lab_sup_an_r,
        "zyg_maj_l": zyg_maj_l,
        "zyg_maj_r": zyg_maj_r,
        "orb_oc_l": orb_oc_l,
        "orb_oc_r": orb_oc_r,
        "orb_oc_l_inner": orb_oc_l_inner,
        "orb_oc_r_inner": orb_oc_r_inner,
        "orb_oris_l": orb_oris_l,
        "orb_oris_u": orb_oris_u,
        "orb_oc_l": orb_oc_l,
        "cor_sup_l": cor_sup_l,
    }

    muscle_names = [
        "bucc_l",
        "bucc_r",
        "masseter_l",
        "masseter_r",
        "temporalis_l",
        "temporalis_r",
        "dep_lab_inf_l",
        "dep_lab_inf_r",
        "dep_ang_or_l",
        "dep_ang_or_r",
        "mentalis_l",
        "mentalis_r",
        "risorius_l",
        "risorius_r",
        "frontalis_l",
        "frontalis_inner_l",
        "frontalis_r",
        "frontalis_inner_r",
        "cor_sup_r",
        "orb_oc_l_outer",
        "orb_oc_r_outer",
        "lev_lab_sup_l",
        "lev_lab_sup_r",
        "lev_lab_sup_an_l",
        "lev_lab_sup_an_r",
        "zyg_maj_l",
        "zyg_maj_r",
        "orb_oc_l",
        "orb_oc_r",
        "orb_oc_l",
        "orb_oc_l_inner",
        "orb_oc_r_inner",
        "orb_oris_l",
        "orb_oris_u",
        "cor_sup_l",
    ]
    todraw = {}
    facet = False

    if "facet" in kwargs and au is not None:
        aus = []
        for i in range(12):
            aus.append(au[i])
        aus.append(au[13])
        aus.append(max(au[12], au[14], au[15], au[18], key=abs))
        aus.append(au[16])
        aus.append(au[17])
        aus.append(au[19])
        au = aus
        facet = True
        del kwargs["facet"]
    if au is None:
        au = np.zeros(20)
    if "all" in kwargs:
        for muscle in muscle_names:
            todraw[muscle] = kwargs["all"]
        del kwargs["all"]
    else:
        for muscle in muscle_names:
            if muscle in kwargs:
                todraw[muscle] = kwargs[muscle]
                del kwargs[muscle]
    for muscle in todraw.keys():
        if todraw[muscle] == "heatmap":
            # muscles[muscle].set_color(get_heat(muscle, au, facet))
            muscles[muscle].set_color(get_heat(muscle, au, facet))
        else:
            muscles[muscle].set_color(todraw[muscle])
        ax.add_patch(muscles[muscle], *args, **kwargs)

    eye_l = plt.Polygon(
        [
            [currx[36], curry[36]],
            [currx[37], curry[37]],
            [currx[38], curry[38]],
            [currx[39], curry[39]],
            [currx[40], curry[40]],
            [currx[41], curry[41]],
        ],
        color="w",
    )

    eye_r = plt.Polygon(
        [
            [currx[42], curry[42]],
            [currx[43], curry[43]],
            [currx[44], curry[44]],
            [currx[45], curry[45]],
            [currx[46], curry[46]],
            [currx[47], curry[47]],
        ],
        color="w",
    )

    mouth = plt.Polygon(
        [
            [currx[60], curry[60]],
            [currx[61], curry[61]],
            [currx[62], curry[62]],
            [currx[63], curry[63]],
            [currx[64], curry[64]],
            [currx[65], curry[65]],
            [currx[66], curry[66]],
            [currx[67], curry[67]],
        ],
        color="w",
    )

    # ax.add_patch(eye_l)
    # ax.add_patch(eye_r)
    # ax.add_patch(mouth)
    return ax


def get_heat(muscle, au, log):
    """Function to create heatmap from au vector

    Args:
        au: vector of action units
        muscle: string representation of a muscle
        boolean: whether the action unit values are on a log scale


    Returns:
        color of muscle according to its au value
    """
    q = sns.color_palette("Blues", 101)
    unit = 0
    aus = {
        "masseter_l": 15,
        "masseter_r": 15,
        "temporalis_l": 15,
        "temporalis_r": 15,
        "dep_lab_inf_l": 14,
        "dep_lab_inf_r": 14,
        "dep_ang_or_l": 10,
        "dep_ang_or_r": 10,
        "mentalis_l": 11,
        "mentalis_r": 11,
        "risorius_l": 12,
        "risorius_r": 12,
        "frontalis_l": 1,
        "frontalis_r": 1,
        "frontalis_inner_l": 0,
        "frontalis_inner_r": 0,
        "cor_sup_l": 2,
        "cor_sup_r": 2,
        "lev_lab_sup_l": 7,
        "lev_lab_sup_r": 7,
        "lev_lab_sup_an_l": 6,
        "lev_lab_sup_an_r": 6,
        "zyg_maj_l": 8,
        "zyg_maj_r": 8,
        "bucc_l": 9,
        "bucc_r": 9,
        "orb_oc_l_outer": 4,
        "orb_oc_r_outer": 4,
        "orb_oc_l": 5,
        "orb_oc_r": 5,
        "orb_oc_l_inner": 16,
        "orb_oc_r_inner": 16,
        "orb_oris_l": 13,
        "orb_oris_u": 13,
    }
    if muscle in aus:
        unit = aus[muscle]
    if log:
        num = int(100 * (1.0 / (1 + 10.0 ** -(au[unit]))))
    else:
        num = int(au[unit] * 20)
    # set alpha (opacity)
    if au[unit] == 0:
        alpha = 0
    else:
        alpha = 0.5
    # color = colors.to_hex(q[num])
    # return str(color)
    color = colors.to_rgba(q[num], alpha=alpha)
    return color


def plot_face(
    model=None,
    au=None,
    vectorfield=None,
    muscles=None,
    ax=None,
    color="k",
    linewidth=1,
    linestyle="-",
    gaze=None,
    *args,
    **kwargs
):
    """Function to plot facesself

    Args:
        model: sklearn PLSRegression instance
        au: vector of action units (same length as model.x_mean_)
        vectorfield: (dict) {'target':target_array,'reference':reference_array}
        muscles: (dict) {'muscle': color}
        ax: matplotlib axis handle
        color: matplotlib color
        linewidth: matplotlib linewidth
        linestyle: matplotlib linestyle
        gaze: array of gaze vectors (len(4))

    Returns:
        ax: plot handle
    """

    if model is None:
        model = load_h5()
    else:
        if not isinstance(model, PLSRegression):
            raise ValueError("make sure that model is a PLSRegression instance")

    if au is None:
        au = np.zeros(len(model.x_mean_))
        warnings.warn(
            "Don't forget to pass an 'au' vector of len(20), "
            "using neutral as default"
        )

    landmarks = predict(au, model)
    currx, curry = [landmarks[x, :] for x in range(2)]

    if ax is None:
        ax = _create_empty_figure()

    if muscles is not None:
        if not isinstance(muscles, dict):
            raise ValueError("muscles must be a dictionary ")
        draw_muscles(currx, curry, ax=ax, au=au, **muscles)

    if gaze is not None and len((gaze)) != 4:
        warnings.warn(
            "Don't forget to pass a 'gaze' vector of len(4), "
            "using neutral as default"
        )
        gaze = None

    draw_lineface(
        currx,
        curry,
        color=color,
        linewidth=linewidth,
        linestyle=linestyle,
        ax=ax,
        gaze=gaze,
        *args,
        **kwargs
    )
    if vectorfield is not None:
        if not isinstance(vectorfield, dict):
            raise ValueError("vectorfield must be a dictionary ")
        if "reference" not in vectorfield:
            raise ValueError("vectorfield must contain 'reference' key")
        if "target" not in vectorfield.keys():
            vectorfield["target"] = landmarks
        draw_vectorfield(ax=ax, **vectorfield)
    ax.set_xlim([25, 172])
    ax.set_ylim((-240, -50))
    ax.axes.get_xaxis().set_visible(False)
    ax.axes.get_yaxis().set_visible(False)
    return ax


def predict(au, model=None):
    """Helper function to predict landmarks from au given a sklearn model

    Args:
        au: vector of action unit intensities
        model: sklearn pls object (uses pretrained model by default)

    Returns:
        landmarks: Array of landmarks (2,68)

    """

    if model is None:
        model = load_h5()
    elif not isinstance(model, PLSRegression):
        raise ValueError("make sure that model is a PLSRegression instance")

    if len(au) != len(model.x_mean_):
        print(au)
        print(len(model.x_mean_))
        raise ValueError("au vector must be len(", len(model.x_mean_), ").")

    if len(au.shape) == 1:
        au = np.reshape(au, (1, -1))

    landmarks = np.reshape(model.predict(au), (2, 68))
    landmarks[1, :] = -1 * landmarks[1, :]  # this might not generalize to other models
    return landmarks


def _create_empty_figure(figsize=(4, 5), xlim=[25, 172], ylim=[-240, -50]):
    """Create an empty figure"""
    plt.figure(figsize=figsize)
    ax = plt.gca()
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    ax.axes.get_xaxis().set_visible(False)
    ax.axes.get_yaxis().set_visible(False)
    return ax



#!/usr/bin/env python3
"""
This script will download the necessary models to the feat path. 

Usage
    python3 download_models.py
"""
from feat.utils import get_resource_path
import wget
import os
import zipfile

if os.path.exists(os.path.join(get_resource_path(), "best_ferModel.pth")):
    print("Fex model already exists; skipping download.")
else:
    print("Downloading FEX emotion model.")
    try:
        fex_emotion_model = "https://github.com/cosanlab/feat/releases/download/v0.1/best_ferModel.pth"
        wget.download(fex_emotion_model, get_resource_path(), bar=None)
    # except:
    #    try:
    #        fex_emotion_model = "https://www.dropbox.com/s/d3yhtsqggqcrjl2/fer_emotion_model.h5?dl=1"
    #        wget.download(fex_emotion_model, get_resource_path(), bar=None)
    except:
        print("FeX emotion model failed to download")

    if os.path.exists(os.path.join(get_resource_path(), "best_ferModel.pth")):
        print("\nFEX emotion model downloaded successfully.\n")
    else:
        print("Something went wrong. Model not found in directory.")

if os.path.exists(os.path.join(get_resource_path(), "lbfmodel.yaml")):
    print("Landmark already exists; skipping download.")
else:
    print("Downloading landmark detection model.")
    try:
        lbfmodel = "https://github.com/cosanlab/feat/releases/download/v0.1/lbfmodel.yaml"
        wget.download(lbfmodel, get_resource_path(), bar=None)
    except:
        try:
            lbfmodel = "https://www.dropbox.com/s/cqune0z1bwf79zy/lbfmodel.yaml?dl=1"
            wget.download(lbfmodel, get_resource_path(), bar=None)
        except:
            print("Landmark model failed to download")

if os.path.exists(os.path.join(get_resource_path(), "align_net.pth")):
    print("\nAU Detection Model Parameters downloaded successfully.\n")
else:
    print("Downloading JAA Net AU Occurence model.")
    try:
        jaanet_params = "https://github.com/cosanlab/feat/releases/download/v0.1/JAANetparams.zip"
        wget.download(jaanet_params, get_resource_path(), bar=None)
        with zipfile.ZipFile(os.path.join(get_resource_path(), "JAANetparams.zip"), 'r') as zip_ref:
            zip_ref.extractall(os.path.join(get_resource_path()))
    except:
        print("JAA parameters failed to download")

    if os.path.exists(os.path.join(get_resource_path(), "align_net.pth")):
        print("\nAU Detection Model Parameters downloaded successfully.\n")
    else:
        print("Something went wrong. Model not found in directory.")

if os.path.exists(os.path.join(get_resource_path(), "DRMLNetParams.pth")):
    print("\nDRML NET model downloaded successfully.\n")
else:
    try:
        # print("Downloading DRML model.")
        drml_model = "https://github.com/cosanlab/feat/releases/download/v0.1/DRMLNetParams.pth"
        wget.download(drml_model, get_resource_path(), bar=None)
        if os.path.exists(os.path.join(get_resource_path(), "DRMLNetParams.pth")):
            print("\nLandmark detection model downloaded successfully.\n")
        else:
            print("Something went wrong. Model not found in directory.")
    except:
        print("DRML model failed to download.")

if os.path.exists(os.path.join(get_resource_path(), "FaceBoxesProd.pth")):
    print("\nFaceBox model downloaded successfully.\n")
else:
    try:
        # print("Downloading nFaceBox model.")
        facebox_model = "https://github.com/cosanlab/feat/releases/download/v0.1/FaceBoxesProd.pth"
        wget.download(facebox_model, get_resource_path(), bar=None)
        if os.path.exists(os.path.join(get_resource_path(), "FaceBoxesProd.pth")):
            print("\nFaceBox model downloaded successfully.\n")
        else:
            print("Something went wrong. Model not found in directory.")
    except:
        print("FaceBox model failed to download.")

if os.path.exists(os.path.join(get_resource_path(), "onet.npy")):
    print("\nMTCNN O Net model downloaded successfully.\n")
else:
    try:
        # print("Downloading MTCNN ONet model.")
        onet_model = "https://github.com/cosanlab/feat/releases/download/v0.1/onet.npy"
        wget.download(onet_model, get_resource_path(), bar=None)
        # if os.path.exists(os.path.join(get_resource_path(), "onet.npy")):
        #     print("\nMTCNN Onet model downloaded successfully.\n")
        # else:
        #     print("Something went wrong. Model not found in directory.")
    except:
        print("MTCNN Onet model failed to download.")

if os.path.exists(os.path.join(get_resource_path(), "pnet.npy")):
    print("\nMTCNN PNet model downloaded successfully.\n")
else:
    try:
        print("Downloading MTCNN pnet model.")
        pnet_model = "https://github.com/cosanlab/feat/releases/download/v0.1/pnet.npy"
        wget.download(pnet_model, get_resource_path(), bar=None)
        if os.path.exists(os.path.join(get_resource_path(), "pnet.npy")):
            print("\nMTCNN pnet model downloaded successfully.\n")
        else:
            print("Something went wrong. Model not found in directory.")
    except:
        print("MTCNN pnet model failed to download.")     

if os.path.exists(os.path.join(get_resource_path(), "rnet.npy")):
    print("\nMTCNN rnet model downloaded successfully.\n")
else:
    try:
        print("Downloading MTCNN rnet model.")
        rnet_model = "https://github.com/cosanlab/feat/releases/download/v0.1/rnet.npy"
        wget.download(rnet_model, get_resource_path(), bar=None)
        if os.path.exists(os.path.join(get_resource_path(), "rnet.npy")):
            print("\nMTCNN rnet model downloaded successfully.\n")
        else:
            print("Something went wrong. Model not found in directory.")
    except:
        print("MTCNN rnet model failed to download.")     

if os.path.exists(os.path.join(get_resource_path(), "mobilenet0.25_Final.pth")):
    print("\nRetinaFace model downloaded successfully.\n")
else:
    try:
        #print("Downloading RetinaFace model.")
        retin_model = "https://github.com/cosanlab/feat/releases/download/v0.1/mobilenet0.25_Final.pth"
        wget.download(retin_model, get_resource_path(), bar=None)
        # if os.path.exists(os.path.join(get_resource_path(), "mobilenet0.25_Final.pth")):
        #     print("\nRetinaFace model downloaded successfully.\n")
        # else:
        #     print("Something went wrong. Model not found in directory.")
    except:
        print("RetinaFace model failed to download.")    

if os.path.exists(os.path.join(get_resource_path(), "mobilenet_224_model_best_gdconv_external.pth.tar")):
    print("\nmobilenet model downloaded successfully.\n")
else:
    try:
        #print("Downloading mobilenet model.")
        mble_model = "https://github.com/cosanlab/feat/releases/download/v0.1/mobilenet_224_model_best_gdconv_external.pth.tar"
        wget.download(mble_model, get_resource_path(), bar=None)
        # if os.path.exists(os.path.join(get_resource_path(), "mobilenet_224_model_best_gdconv_external.pth.tar")):
        #     print("\nmobilenet model downloaded successfully.\n")
        # else:
        #     print("Something went wrong. Model not found in directory.")
    except:
        print("mobilenet model failed to download.")    

if os.path.exists(os.path.join(get_resource_path(), "pfld_model_best.pth.tar")):
    print("\nPFLD model downloaded successfully.\n")
else:
    try:
        #print("Downloading PFLD model.")
        pfld_model = "https://github.com/cosanlab/feat/releases/download/v0.1/pfld_model_best.pth.tar"
        wget.download(pfld_model, get_resource_path(), bar=None)
        # if os.path.exists(os.path.join(get_resource_path(), "pfld_model_best.pth.tar")):
        #     print("\nPFLD model downloaded successfully.\n")
        # else:
        #     print("Something went wrong. Model not found in directory.")
    except:
        print("PFLD model failed to download.")    

if os.path.exists(os.path.join(get_resource_path(), "mobilefacenet_model_best.pth.tar")):
    print("\nMobileFaceNet model downloaded successfully.\n")
else:
    try:
        #print("Downloading MobileFaceNet model.")
        mbfa_model = "https://github.com/cosanlab/feat/releases/download/v0.1/mobilefacenet_model_best.pth.tar"
        wget.download(mbfa_model, get_resource_path(), bar=None)
        # if os.path.exists(os.path.join(get_resource_path(), "mobilefacenet_model_best.pth.tar")):
        #     print("\nMobileFaceNet model downloaded successfully.\n")
        # else:
        #     print("Something went wrong. Model not found in directory.")
    except:
        print("MobileFaceNet model failed to download.")   

if os.path.exists(os.path.join(get_resource_path(), "ResMaskNet_Z_resmasking_dropout1_rot30.pth")):
    print("\nResidualMaskingNetwork model downloaded successfully.\n")
else:
    try:
        resmasknetmodel = "https://github.com/cosanlab/feat/releases/download/v0.1/ResMaskNet_Z_resmasking_dropout1_rot30.pth"
        wget.download(resmasknetmodel, get_resource_path(), bar=None)
    except:
        print("ResidualMaskingNetwork model failed to download.")   


#!/usr/bin/env python3
'''
Multithreaded emotion & facial features detector from videos.

Usage:
    python3 detect_fex.py
    Detect facial landmarks, emotions, and action units from video file or see a demo with webcam.

Arguments:
    inputFname: str
            Path to video file to process. If no video file is passed, then provide demo with camera.
    ouputFname: str
            Path to output file to write to.
    skip_frames: int, default = 1
            Process every n-th frame for faster speed. 
    n_jobs: int, default=1
            The number of jobs to use for the computation. Default uses 1 core. 
            -1 means using all processors. 

Outputs:
    If outputFname is specified, emotion predictions, face and landmark detectionresults will be saved as a csv file.
Keyboard shortcuts:
    ESC - exit
'''

if __name__ == '__main__':
    import sys
    import argparse
    import os
    print(__doc__)
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--inputFname", help="str, Path to video.", type=str, default=0)
    parser.add_argument("-o", "--outputFname", help="str, Path to output file.", type=str)
    parser.add_argument("-s", "--skip_frames", help="int, Process every n-th frame.", type=int, default=1)
    parser.add_argument("-n", "--n_jobs", help="int, Number of processors to use", type=int, default = 1)
    parser.add_argument("-v", "--verbose", help="int, Number of processors to use", default = False)
    args = parser.parse_args()

    if not args.outputFname:
        if args.inputFname==0:
            args.outputFname = None
        else:
            args.outputFname = os.path.splitext(args.inputFname)[0]+".csv"
    print(f"Processing video {args.inputFname}")        
    print(f"Outputs will be saved to {args.outputFname}")

from collections import deque
from multiprocessing.pool import ThreadPool
import tensorflow as tf
from tensorflow.python.keras import optimizers, models
import numpy as np, pandas as pd
from PIL import Image, ImageDraw
import cv2 as cv
from feat.utils import get_resource_path

def face_rect_to_coords(rectangle):
    """
    Takes in a (x, y, w, h) array and transforms it into (x, y, x2, y2)
    """
    return [rectangle[0], rectangle[1], rectangle[0]+rectangle[2], rectangle[1]+rectangle[3]]

# load pre trained emotion model
print("Loading FEX DCNN model.")
dcnn_model_path = os.path.join(get_resource_path(),'fer_aug_model.h5')
if not os.path.exists(dcnn_model_path):
    print("Emotion prediction model not found. Please run download_models.py.")
model = models.load_model(dcnn_model_path) # Load model to use.    
(_, img_w, img_h, img_c) = model.layers[0].input_shape # model input shape.
mapper = {0:'anger', 1:'disgust', 2:'fear', 3:'happiness', 4: 'sadness', 5: 'surprise', 6: 'neutral'}
emotion_columns = [key for key in mapper.values()]

print("Loading Face Detection model.")
face_detection_model_path = cv.data.haarcascades + "haarcascade_frontalface_alt.xml"
if not os.path.exists(face_detection_model_path):
    print("Face detection model not found. Check haarcascade_frontalface_alt.xml exists in your opencv installation (cv.data).")
face_cascade = cv.CascadeClassifier(face_detection_model_path)
facebox_columns = ["facebox_x", "facebox_y", "facebox_w", "facebox_h"]
facebox_empty = np.empty((1,4))
facebox_empty[:] = np.nan
empty_facebox = pd.DataFrame(facebox_empty, columns = facebox_columns)

print("Loading Face Landmark model.")
face_landmark = cv.face.createFacemarkLBF()
landmark_model_path = os.path.join(get_resource_path(), 'lbfmodel.yaml')
if not os.path.exists(landmark_model_path):
    print("Face landmark model not found. Please run download_models.py.")
face_landmark.loadModel(landmark_model_path)
landmark_columns = np.array([(f'x_{i}',f'y_{i}') for i in range(68)]).reshape(1,136)[0].tolist()
landmark_empty = np.empty((1,136))
landmark_empty[:] = np.nan
empty_landmark = pd.DataFrame(landmark_empty, columns = landmark_columns)

# create empty df for predictions 
predictions = np.empty((1, len(mapper)))
predictions[:] = np.nan
empty_df = pd.DataFrame(predictions, columns = mapper.values())

frame_columns = ["frame"]
init_df = pd.DataFrame(columns=frame_columns+emotion_columns+facebox_columns+landmark_columns)
if args.outputFname:
    init_df.to_csv(args.outputFname, index=False, header=True)

def process_frame(frame, counter, input_image_size = (48,48), mapper=mapper):
    """
    Takes a frame from OpenCV and prepares it as a tensor to be predicted by model. 

    frame: Image input. 
    counter: Frame number. 
    """
    try:
        # change image to grayscale
        grayscale_image = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
        # find faces
        detected_faces = face_cascade.detectMultiScale(grayscale_image)
        # detect landmarks
        ok, landmarks = face_landmark.fit(grayscale_image, detected_faces)
        landmarks_df = pd.DataFrame(landmarks[0][0].reshape(1, 136), columns = landmark_columns, index=[counter])
        # Use Tiankang's colde to align the faces to the center 

        # crop just the face area
        if len(detected_faces)>0:
            facebox_df = pd.DataFrame([detected_faces[0]], columns = facebox_columns, index=[counter])

            grayscale_cropped_face = Image.fromarray(grayscale_image).crop(face_rect_to_coords(detected_faces[0]))
            # resize face to newsize 48 x 48
            # print("resizeface", grayscale_cropped_face.shape, img_w, img_h, img_c) 
            grayscale_cropped_resized_face = grayscale_cropped_face.resize(input_image_size)
            # reshape to put in model
            grayscale_cropped_resized_reshaped_face = np.array(grayscale_cropped_resized_face).reshape(1, img_w, img_h, img_c)
            # normalize
            normalize_grayscale_cropped_resized_reshaped_face = grayscale_cropped_resized_reshaped_face/255.
            # make tensor
            tensor_img = tf.convert_to_tensor(normalize_grayscale_cropped_resized_reshaped_face)
            # make predictions
            predictions = model.predict(tensor_img)
            emotion_df = pd.DataFrame(predictions, columns = mapper.values(), index=[counter])
            return emotion_df, frame, facebox_df.reindex(index=[counter]), landmarks_df.reindex(index=[counter])
    except:
        return empty_df.reindex(index=[counter]), frame, empty_facebox.reindex(index=[counter]), empty_landmark.reindex(index=[counter])

if __name__ == '__main__':
    # Setup.
    cap = cv.VideoCapture(args.inputFname)
    fps = cap.get(cv.CAP_PROP_FPS) # get frames per second. 
    
    # Determine whether to use multiprocessing.
    if args.n_jobs==-1:
        thread_num = cv.getNumberOfCPUs() # get available cpus
    else: 
        thread_num = args.n_jobs
    pool = ThreadPool(processes=thread_num)

    pending_task = deque()

    counter = 0
    processed_frames = 0
    detected_faces = []
    print("Processing video.")
    while True:
        # Consume the queue.
        while len(pending_task) > 0 and pending_task[0].ready():
            emotion_df, frame, facebox_df, landmarks_df = pending_task.popleft().get()
            df = pd.concat([emotion_df, facebox_df, landmarks_df], axis=1)
            if args.verbose:
                print(emotion_df)
            # Save to output file.
            if args.outputFname:
                df.to_csv(args.outputFname, index=True, header=False, mode='a')
            processed_frames = processed_frames + 1
            # if we are in demo mode.
            if args.inputFname == 0:           
                # if a face exists
                detected_faces = facebox_df.values
                if not np.isnan(detected_faces[0]).any():
                    # draw rectangle on face.
                    (x, y, xw, yh) = face_rect_to_coords(detected_faces[0])
                    cv.rectangle(frame, (x,y), (xw, yh), (0,255,0))
                    labels = [key+": "+str(np.round(df.loc[:, key].values[0],2)) for key in mapper.values()]
                    y0, dy = 50, 30
                    for i, line in enumerate(labels):
                        y = y0 + i*dy
                        cv.putText(frame, line, (30,y), cv.FONT_HERSHEY_SIMPLEX, 1, (0,255,0), 3)
                else:
                    cv.putText(frame, 'No Face Found', (30,30), cv.FONT_HERSHEY_SIMPLEX, 1, (0,255,0), 3)
                # Show preview.
                cv.imshow('Emotion detector demo', frame)
                if cv.waitKey(1) == 27:
                    break

        # Populate the queue.
        if len(pending_task) < thread_num:
            frame_got, frame = cap.read()
            # Process at every seconds. 
            if counter%args.skip_frames == 0:
                if frame_got:
                    task = pool.apply_async(process_frame, (frame.copy(), counter))
                    pending_task.append(task)
            counter = counter + 1

        if not frame_got:
            break
cap.release() 
cv.destroyAllWindows()


from __future__ import division

"""
    FEAT Utils Class
    ==========================================
    read_facet: read in iMotions-FACET formatted files
    read_affdex: read in iMotions-affdex formatted files
    read_affectiva: read in affectiva-api formatted files
    read_openface: read in openface formatted files

"""

__all__ = [
    "get_resource_path",
    "load_pickled_model",
    "read_facet",
    "read_affdex",
    "read_affectiva",
    "read_openface",
    "softmax",
    "registration",
    "neutral",
    "load_h5",
]
__author__ = ["Jin Hyun Cheong"]


import os, math, pywt, pickle, h5py
from sklearn.cross_decomposition import PLSRegression
import numpy as np, pandas as pd
from scipy import signal
from scipy.integrate import simps
import feat
import cv2
import math
from PIL import Image
import torch

""" DEFINE IMPORTANT VARIABLES """
# FEAT columns
FEAT_EMOTION_MAPPER = {
    0: "anger",
    1: "disgust",
    2: "fear",
    3: "happiness",
    4: "sadness",
    5: "surprise",
    6: "neutral",
}
FEAT_EMOTION_COLUMNS = [
    "anger",
    "disgust",
    "fear",
    "happiness",
    "sadness",
    "surprise",
    "neutral",
]
FEAT_FACEBOX_COLUMNS = [
    "FaceRectX",
    "FaceRectY",
    "FaceRectWidth",
    "FaceRectHeight",
    "FaceScore",
]
# FEAT_FACEBOX_COLUMNS = ['FaceRectX1','FaceRectY1','FaceRectX2','FaceRectY2']
FEAT_TIME_COLUMNS = ["frame"]

# FACET columns
FACET_EMOTION_COLUMNS = [
    "Joy",
    "Anger",
    "Surprise",
    "Fear",
    "Contempt",
    "Disgust",
    "Sadness",
    "Confusion",
    "Frustration",
    "Neutral",
    "Positive",
    "Negative",
]
FACET_FACEBOX_COLUMNS = ["FaceRectX", "FaceRectY", "FaceRectWidth", "FaceRectHeight"]
FACET_TIME_COLUMNS = ["Timestamp", "MediaTime", "FrameNo", "FrameTime"]
FACET_FACEPOSE_COLUMNS = ["Pitch", "Roll", "Yaw"]
FACET_DESIGN_COLUMNS = ["StimulusName", "SlideType", "EventSource", "Annotation"]

# OpenFace columns
landmark_length = 68
openface_2d_landmark_columns = [f"x_{i}" for i in range(landmark_length)] + [
    f"y_{i}" for i in range(landmark_length)
]
openface_3d_landmark_columns = (
    [f"X_{i}" for i in range(landmark_length)]
    + [f"Y_{i}" for i in range(landmark_length)]
    + [f"Z_{i}" for i in range(landmark_length)]
)

jaanet_AU_list = [1, 2, 4, 6, 7, 10, 12, 14, 15, 17, 23, 24]
jaanet_AU_presence = [f"AU" + str(i).zfill(2) for i in jaanet_AU_list]
jaanet_AU_presence.sort()

openface_AU_list = [1, 2, 4, 5, 6, 7, 9, 10, 12, 14, 15, 17, 20, 23, 25, 26, 45]
openface_AU_intensity = [f"AU" + str(i).zfill(2) + "_r" for i in openface_AU_list]
openface_AU_presence = [f"AU" + str(i).zfill(2) + "_c" for i in openface_AU_list + [28]]
openface_AU_presence.sort()
openface_AU_columns = openface_AU_intensity + openface_AU_presence
openface_time_columns = ["frame", "timestamp"]
openface_gaze_columns = [
    "gaze_0_x",
    "gaze_0_y",
    "gaze_0_z",
    "gaze_1_x",
    "gaze_1_y",
    "gaze_1_z",
]
openface_facepose_columns = [
    "pose_Tx",
    "pose_Ty",
    "pose_Tz",
    "pose_Rx",
    "pose_Ry",
    "pose_Rz",
]
OPENFACE_ORIG_COLUMNS = (
    openface_time_columns
    + ["confidence", "success"]
    + openface_gaze_columns
    + openface_facepose_columns
    + openface_2d_landmark_columns
    + openface_3d_landmark_columns
    + [
        "p_scale",
        "p_rx",
        "p_ry",
        "p_rz",
        "p_tx",
        "p_ty",
        "p_0",
        "p_1",
        "p_2",
        "p_3",
        "p_4",
        "p_5",
        "p_6",
        "p_7",
        "p_8",
        "p_9",
        "p_10",
        "p_11",
        "p_12",
        "p_13",
        "p_14",
        "p_15",
        "p_16",
        "p_17",
        "p_18",
        "p_19",
        "p_20",
        "p_21",
        "p_22",
        "p_23",
        "p_24",
        "p_25",
        "p_26",
        "p_27",
        "p_28",
        "p_29",
        "p_30",
        "p_31",
        "p_32",
        "p_33",
    ]
    + openface_AU_columns
)


def face_rect_to_coords(rectangle):
    """
    Takes in a (x, y, w, h) array and transforms it into (x, y, x2, y2)
    """
    return [
        rectangle[0],
        rectangle[1],
        rectangle[0] + rectangle[2],
        rectangle[1] + rectangle[3],
    ]


def get_resource_path():
    """ Get path to feat resource directory. """
    return os.path.join(feat.__path__[0], "resources")  # points to the package folder.
    # return ("F:/feat/feat/") # points to the package folder.
    # return os.path.join(os.path.dirname(__file__), 'resources')


def load_pickled_model(file_name="pls_python27.pkl"):
    """Load the pickled PLS model for plotting.

    Args:
        file_name ([type], optional): Specify model to load. Defaults to  pls_python27.pkl.

    Returns:
        model: PLS model
    """
    file_name = os.path.join(get_resource_path(), file_name)
    try:
        with open(file_name, "rb") as f:
            model = pickle.load(f)
    except UnicodeDecodeError as e:
        with open(file_name, "rb") as f:
            model = pickle.load(f, encoding="latin1")
    except Exception as e:
        print("Unable to load data ", file_name, ":", e)
        raise
    return model


def load_h5(file_name="blue.h5"):
    """Load the h5 PLS model for plotting.

    Args:
        file_name (str, optional): Specify model to load.. Defaults to 'blue.h5'.

    Returns:
        model: PLS model
    """
    try:
        hf = h5py.File(os.path.join(get_resource_path(), file_name), "r")
        d1 = hf.get("coef")
        d2 = hf.get("x_mean")
        d3 = hf.get("y_mean")
        d4 = hf.get("x_std")
        model = PLSRegression(len(d1))
        model.coef_ = np.array(d1)
        model.x_mean_ = np.array(d2)
        model.y_mean_ = np.array(d3)
        model.x_std_ = np.array(d4)
        hf.close()
    except Exception as e:
        print("Unable to load data ", file_name, ":", e)
    return model


def read_feat(fexfile):
    """This function reads files extracted using the Detector from the Feat package.

    Args:
        fexfile: [description]
    """
    d = pd.read_csv(fexfile)
    return feat.Fex(
        d,
        filename=fexfile,
        au_columns=jaanet_AU_presence,
        emotion_columns=FEAT_EMOTION_COLUMNS,
        landmark_columns=openface_2d_landmark_columns,
        facebox_columns=FEAT_FACEBOX_COLUMNS,
        time_columns=FEAT_TIME_COLUMNS,
        detector="Feat",
    )


def read_facet(facetfile, features=None, raw=False, sampling_freq=None):
    """This function reads in an iMotions-FACET exported facial expression file.

    Args:
        facetfile: iMotions-FACET file. Files from iMotions 5, 6, and 7 have been tested and supported
        features: If a list of iMotion-FACET column names are passed, those are returned.
                Otherwise, default columns are returned in the following format:['Timestamp','FaceRectX','FaceRectY','FaceRectWidth','FaceRectHeight',
                'Joy','Anger','Surprise','Fear','Contempt', 'Disgust','Sadness','Confusion','Frustration',
                'Neutral','Positive','Negative','AU1','AU2', 'AU4','AU5','AU6','AU7','AU9','AU10',
                'AU12','AU14','AU15','AU17','AU18','AU20', 'AU23','AU24','AU25','AU26','AU28','AU43',
                'Yaw', 'Pitch', 'Roll'].
                Note that these column names are different from the original files which has ' Evidence', ' Degrees' appended to each column.
        raw (default=False): Set to True to return all columns without processing.
        sampling_freq: sampling frequency to pass to Fex
    Returns:
        dataframe of processed facial expressions
    """
    # Check iMotions Version
    versionstr = ""
    try:
        with open(facetfile, "r") as f:
            studyname = f.readline().replace("\t", "").replace("\n", "")
            studydate = f.readline().replace("\t", "").replace("\n", "")
            versionstr = f.readline().replace("\t", "").replace("\n", "")
        versionnum = int(versionstr.split(" ")[-1].split(".")[0])
    except:
        raise TypeError(
            "Cannot infer version of iMotions-FACET file. Check to make sure this is the raw iMotions-FACET file."
        )

    d = pd.read_csv(facetfile, skiprows=5, sep="\t")
    # Check if features argument is passed and return only those features, else return all columns
    if isinstance(features, list):
        try:
            d = d[features]
            if raw:
                return feat.Fex(d, filename=facetfile)
        except:
            raise KeyError([features, "not in facetfile"])
    elif isinstance(features, type(None)):
        if raw:
            return feat.Fex(d, filename=facetfile)
        else:
            fex_columns = [
                col.replace(" Evidence", "").replace(" Degrees", "")
                for col in d.columns
                if "Evidence" in col or "Degrees" in col
            ]
            # Remove Intensity as this has been deprecated
            cols2drop = [col for col in d.columns if "Intensity" in col]
            d = d.drop(columns=cols2drop)
            d.columns = [col.replace(" Evidence", "") for col in d.columns]
            d.columns = [col.replace(" Degrees", "") for col in d.columns]
            d.columns = [col.replace(" ", "") for col in d.columns]
            # d._metadata = fex_columns
    au_columns = [col for col in d.columns if "AU" in col]
    return feat.Fex(
        d,
        filename=facetfile,
        au_columns=au_columns,
        emotion_columns=FACET_EMOTION_COLUMNS,
        facebox_columns=FACET_FACEBOX_COLUMNS,
        facepose_columns=FACET_FACEPOSE_COLUMNS,
        time_columns=FACET_TIME_COLUMNS,
        design_columns=FACET_DESIGN_COLUMNS,
        detector="FACET",
        sampling_freq=sampling_freq,
    )


def read_openface(openfacefile, features=None):
    """
    This function reads in an OpenFace exported facial expression file.
    Args:
        features: If a list of column names are passed, those are returned. Otherwise, default returns the following features:
        ['frame', 'timestamp', 'confidence', 'success', 'gaze_0_x',
       'gaze_0_y', 'gaze_0_z', 'gaze_1_x', 'gaze_1_y', 'gaze_1_z',
       'pose_Tx', 'pose_Ty', 'pose_Tz', 'pose_Rx', 'pose_Ry', 'pose_Rz',
       'x_0', 'x_1', 'x_2', 'x_3', 'x_4', 'x_5', 'x_6', 'x_7', 'x_8',
       'x_9', 'x_10', 'x_11', 'x_12', 'x_13', 'x_14', 'x_15', 'x_16',
       'x_17', 'x_18', 'x_19', 'x_20', 'x_21', 'x_22', 'x_23', 'x_24',
       'x_25', 'x_26', 'x_27', 'x_28', 'x_29', 'x_30', 'x_31', 'x_32',
       'x_33', 'x_34', 'x_35', 'x_36', 'x_37', 'x_38', 'x_39', 'x_40',
       'x_41', 'x_42', 'x_43', 'x_44', 'x_45', 'x_46', 'x_47', 'x_48',
       'x_49', 'x_50', 'x_51', 'x_52', 'x_53', 'x_54', 'x_55', 'x_56',
       'x_57', 'x_58', 'x_59', 'x_60', 'x_61', 'x_62', 'x_63', 'x_64',
       'x_65', 'x_66', 'x_67', 'y_0', 'y_1', 'y_2', 'y_3', 'y_4', 'y_5',
       'y_6', 'y_7', 'y_8', 'y_9', 'y_10', 'y_11', 'y_12', 'y_13', 'y_14',
       'y_15', 'y_16', 'y_17', 'y_18', 'y_19', 'y_20', 'y_21', 'y_22',
       'y_23', 'y_24', 'y_25', 'y_26', 'y_27', 'y_28', 'y_29', 'y_30',
       'y_31', 'y_32', 'y_33', 'y_34', 'y_35', 'y_36', 'y_37', 'y_38',
       'y_39', 'y_40', 'y_41', 'y_42', 'y_43', 'y_44', 'y_45', 'y_46',
       'y_47', 'y_48', 'y_49', 'y_50', 'y_51', 'y_52', 'y_53', 'y_54',
       'y_55', 'y_56', 'y_57', 'y_58', 'y_59', 'y_60', 'y_61', 'y_62',
       'y_63', 'y_64', 'y_65', 'y_66', 'y_67', 'X_0', 'X_1', 'X_2', 'X_3',
       'X_4', 'X_5', 'X_6', 'X_7', 'X_8', 'X_9', 'X_10', 'X_11', 'X_12',
       'X_13', 'X_14', 'X_15', 'X_16', 'X_17', 'X_18', 'X_19', 'X_20',
       'X_21', 'X_22', 'X_23', 'X_24', 'X_25', 'X_26', 'X_27', 'X_28',
       'X_29', 'X_30', 'X_31', 'X_32', 'X_33', 'X_34', 'X_35', 'X_36',
       'X_37', 'X_38', 'X_39', 'X_40', 'X_41', 'X_42', 'X_43', 'X_44',
       'X_45', 'X_46', 'X_47', 'X_48', 'X_49', 'X_50', 'X_51', 'X_52',
       'X_53', 'X_54', 'X_55', 'X_56', 'X_57', 'X_58', 'X_59', 'X_60',
       'X_61', 'X_62', 'X_63', 'X_64', 'X_65', 'X_66', 'X_67', 'Y_0',
       'Y_1', 'Y_2', 'Y_3', 'Y_4', 'Y_5', 'Y_6', 'Y_7', 'Y_8', 'Y_9',
       'Y_10', 'Y_11', 'Y_12', 'Y_13', 'Y_14', 'Y_15', 'Y_16', 'Y_17',
       'Y_18', 'Y_19', 'Y_20', 'Y_21', 'Y_22', 'Y_23', 'Y_24', 'Y_25',
       'Y_26', 'Y_27', 'Y_28', 'Y_29', 'Y_30', 'Y_31', 'Y_32', 'Y_33',
       'Y_34', 'Y_35', 'Y_36', 'Y_37', 'Y_38', 'Y_39', 'Y_40', 'Y_41',
       'Y_42', 'Y_43', 'Y_44', 'Y_45', 'Y_46', 'Y_47', 'Y_48', 'Y_49',
       'Y_50', 'Y_51', 'Y_52', 'Y_53', 'Y_54', 'Y_55', 'Y_56', 'Y_57',
       'Y_58', 'Y_59', 'Y_60', 'Y_61', 'Y_62', 'Y_63', 'Y_64', 'Y_65',
       'Y_66', 'Y_67', 'Z_0', 'Z_1', 'Z_2', 'Z_3', 'Z_4', 'Z_5', 'Z_6',
       'Z_7', 'Z_8', 'Z_9', 'Z_10', 'Z_11', 'Z_12', 'Z_13', 'Z_14', 'Z_15',
       'Z_16', 'Z_17', 'Z_18', 'Z_19', 'Z_20', 'Z_21', 'Z_22', 'Z_23',
       'Z_24', 'Z_25', 'Z_26', 'Z_27', 'Z_28', 'Z_29', 'Z_30', 'Z_31',
       'Z_32', 'Z_33', 'Z_34', 'Z_35', 'Z_36', 'Z_37', 'Z_38', 'Z_39',
       'Z_40', 'Z_41', 'Z_42', 'Z_43', 'Z_44', 'Z_45', 'Z_46', 'Z_47',
       'Z_48', 'Z_49', 'Z_50', 'Z_51', 'Z_52', 'Z_53', 'Z_54', 'Z_55',
       'Z_56', 'Z_57', 'Z_58', 'Z_59', 'Z_60', 'Z_61', 'Z_62', 'Z_63',
       'Z_64', 'Z_65', 'Z_66', 'Z_67', 'p_scale', 'p_rx', 'p_ry', 'p_rz',
       'p_tx', 'p_ty', 'p_0', 'p_1', 'p_2', 'p_3', 'p_4', 'p_5', 'p_6',
       'p_7', 'p_8', 'p_9', 'p_10', 'p_11', 'p_12', 'p_13', 'p_14', 'p_15',
       'p_16', 'p_17', 'p_18', 'p_19', 'p_20', 'p_21', 'p_22', 'p_23',
       'p_24', 'p_25', 'p_26', 'p_27', 'p_28', 'p_29', 'p_30', 'p_31',
       'p_32', 'p_33', 'AU01_r', 'AU02_r', 'AU04_r', 'AU05_r', 'AU06_r',
       'AU07_r', 'AU09_r', 'AU10_r', 'AU12_r', 'AU14_r', 'AU15_r',
       'AU17_r', 'AU20_r', 'AU23_r', 'AU25_r', 'AU26_r', 'AU45_r',
       'AU01_c', 'AU02_c', 'AU04_c', 'AU05_c', 'AU06_c', 'AU07_c',
       'AU09_c', 'AU10_c', 'AU12_c', 'AU14_c', 'AU15_c', 'AU17_c',
       'AU20_c', 'AU23_c', 'AU25_c', 'AU26_c', 'AU28_c', 'AU45_c']

    Returns:
        dataframe of processed facial expressions

    """
    d = pd.read_csv(openfacefile, sep=",")
    d.columns = d.columns.str.strip(" ")
    # Check if features argument is passed and return only those features, else return basic emotion/AU features
    if isinstance(features, list):
        try:
            d = d[features]
        except:
            raise KeyError([features, "not in openfacefile"])
    elif isinstance(features, type(None)):
        features = OPENFACE_ORIG_COLUMNS
        try:
            d = d[features]
        except:
            pass
    return feat.Fex(
        d,
        filename=openfacefile,
        au_columns=openface_AU_columns,
        emotion_columns=None,
        facebox_columns=None,
        landmark_columns=openface_2d_landmark_columns,
        facepose_columns=openface_facepose_columns,
        gaze_columns=openface_gaze_columns,
        time_columns=openface_time_columns,
        detector="OpenFace",
    )


def read_affectiva(affectivafile, orig_cols=False):
    """
    This function reads in affectiva file processed through
    the https://github.com/cosanlab/affectiva-api-app.
    Args:
        affectivafile: file to read
        orig_cols: If True, convert original colnames to FACS names
    """
    d = pd.read_json(affectivafile, lines=True)
    rep_dict = {
        "anger": "Anger",
        "attention": "Attention",
        "contempt": "Contempt",
        "disgust": "Disgust",
        "engagement": "Engagement",
        "fear": "Fear",
        "joy": "Joy",
        "sadness": "Sadness",
        "smirk": "Smirk",
        "surprise": "Surprise",
        "valence": "Valence",
        "browFurrow": "AU04",
        "smile": "AU12",
        "browRaise": "AU02",
        "cheekRaise": "AU06",
        "chinRaise": "AU17",
        "dimpler": "AU14",
        "eyeClosure": "AU43",
        "eyeWiden": "AU05",
        "innerBrowRaise": "AU01",
        "jawDrop": "AU26",
        "lidTighten": "AU07",
        "lipCornerDepressor": "AU15",
        "lipPress": "AU24",
        "lipPucker": "AU18",
        "lipStretch": "AU20",
        "lipSuck": "AU28",
        "mouthOpen": "AU25",
        "noseWrinkle": "AU09",
        "upperLipRaise": "AU10",
    }
    affectiva_au_columns = [col for col in rep_dict.values() if "AU" in col]
    affectiva_emotion_columns = list(set(rep_dict.values()) - set(affectiva_au_columns))
    if not orig_cols:
        new_cols = []
        for col in d.columns:
            try:
                new_cols.append(rep_dict[col])
            except:
                new_cols.append(col)
        d.columns = new_cols
    return feat.Fex(
        d,
        filename=affectivafile,
        au_columns=affectiva_au_columns,
        emotion_columns=affectiva_emotion_columns,
        detector="Affectiva",
    )


def wavelet(freq, num_cyc=3, sampling_freq=30.0):
    """Create a complex Morlet wavelet by windowing a cosine function by a
    Gaussian. All formulae taken from Cohen, 2014 Chaps 12 + 13

    Args:
        freq: (float) desired frequence of wavelet
        num_cyc: (float) number of wavelet cycles/gaussian taper. Note that
                 smaller cycles give greater temporal precision and that larger
                 values give greater frequency precision; (default: 3)
        sampling_freq: (float) sampling frequency of original signal.

    Returns:
        wav: (ndarray) complex wavelet
    """
    dur = (1 / freq) * num_cyc
    time = np.arange(-dur, dur, 1.0 / sampling_freq)

    # Cosine component
    sin = np.exp(2 * np.pi * 1j * freq * time)

    # Gaussian component
    sd = num_cyc / (2 * np.pi * freq)  # standard deviation
    gaus = np.exp(-(time ** 2.0) / (2.0 * sd ** 2.0))

    return sin * gaus


def calc_hist_auc(vals, hist_range=None):
    """
    This function follows the bag of temporal feature analysis as described in
    Bartlett, M. S., Littlewort, G. C., Frank, M. G., & Lee, K. (2014).
    Automatic decoding of facial movements reveals deceptive pain expressions.
    Current Biology, 24(7), 738-743.
    The function receives convolved data, squares the values,
    finds 0 crossings to calculate the AUC(area under the curve)
    and generates a 6 exponentially-spaced-bin histogram for each data.

    Args:

    Returns:


    """
    # Square values
    vals = [elem ** 2 if elem > 0 else -1 * elem ** 2 for elem in vals]
    # Get 0 crossings
    crossings = np.where(np.diff(np.sign(vals)))[0]
    pos, neg = [], []
    for i in range(len(crossings)):
        if i == 0:
            cross = vals[: crossings[i]]
        elif i == len(crossings) - 1:
            cross = vals[crossings[i] :]
        else:
            cross = vals[crossings[i] : crossings[i + 1]]
        if cross:
            auc = simps(cross)
            if auc > 0:
                pos.append(auc)
            elif auc < 0:
                neg.append(np.abs(auc))
    if not hist_range:
        hist_range = np.logspace(0, 5, 7)  # bartlett 10**0~ 10**5

    out = pd.Series(
        np.hstack([np.histogram(pos, hist_range)[0], np.histogram(neg, hist_range)[0]])
    )
    return out


def softmax(x):
    """
    Softmax function to change log likelihood evidence values to probabilities.
    Use with Evidence values from FACET.

    Args:
        x: value to softmax
    """
    return 1.0 / (1 + 10.0 ** -(x))


### Functions for face registration ###
neutral = np.array(
    [
        [37.514994071403564, 118.99554304280198],
        [38.347467261268164, 135.93119298564565],
        [40.77550102890035, 152.83280452855092],
        [44.109285817364565, 169.1279402172728],
        [49.982831719005134, 184.53328583541997],
        [59.18894827224358, 198.01613609507382],
        [70.41509055106278, 209.2829929016551],
        [83.65962787515429, 217.8257797774197],
        [98.6747861407431, 220.00636721799012],
        [113.36502269321642, 217.35622273914575],
        [126.09720342342395, 208.61554139570768],
        [137.37278216681938, 197.26636201144768],
        [146.15109522110836, 183.95054534968338],
        [151.7203254679301, 168.70328047716666],
        [154.90171533762026, 152.54959546525106],
        [157.01705755745184, 136.0791940145902],
        [157.81240022435486, 119.28714731581948],
        [45.87342275811805, 109.05187535227455],
        [53.83702202368147, 101.43275042887998],
        [65.61231530318975, 99.44649503101734],
        [77.49003981781006, 101.34627038289048],
        [88.31833069100318, 105.66229287035226],
        [108.80512997829634, 105.18583248508406],
        [120.180518838434, 100.84850879934683],
        [131.6712265255873, 99.22426247038426],
        [142.8040873694427, 101.39810664193074],
        [150.0927107560624, 108.74640334130906],
        [98.93955016899183, 117.16643104438056],
        [99.01139789533919, 128.44882090731443],
        [99.09059391932496, 139.71335180079416],
        [99.22411612922204, 151.32196734885628],
        [85.97238779261347, 158.19140086045783],
        [92.2064468619444, 160.61659751751895],
        [98.67862473703293, 162.56437315387998],
        [105.26853264390792, 160.62509055875418],
        [111.14227856687825, 158.32687793454852],
        [59.22833204018989, 118.63189570941351],
        [66.08746862218415, 114.39263501569359],
        [74.66886627073309, 114.59919005618073],
        [81.80683310969295, 120.00819188630955],
        [74.34426159695313, 121.7055175900537],
        [65.7237769427475, 121.82223252349223],
        [114.7522889881524, 119.90654628204749],
        [122.29832379941683, 114.26349216485505],
        [130.61954432603773, 114.38399042631573],
        [137.03708638863128, 118.48489574810866],
        [131.21518765419418, 121.51217888800802],
        [122.97461037812238, 121.56526096419978],
        [75.39827955150834, 179.4070640864827],
        [84.55991401346533, 176.2145796986134],
        [92.90235587470646, 174.4243211955652],
        [98.56534031739243, 176.0653659731581],
        [104.97777372929698, 174.45766843787231],
        [113.1125749468363, 176.39970964033202],
        [121.19973608809934, 179.19790184992027],
        [113.16310623913299, 185.69051008752652],
        [105.26365304952049, 188.31443911070232],
        [98.41771871303214, 188.9656394139811],
        [92.2240282658315, 188.38538897373022],
        [84.05109731022314, 185.74954657843966],
        [79.18422925303048, 179.8065722186372],
        [92.7172317110304, 179.5201781895618],
        [98.52973444977067, 180.1630365496041],
        [105.05932172975814, 179.42368920844928],
        [117.43706438358437, 179.7109259873213],
        [104.90869094557993, 180.32984591574524],
        [98.35933953480642, 181.15981769637827],
        [92.49485174856926, 180.48994809345996],
    ]
)


def registration(face_lms, neutral=neutral, method="fullface"):
    """
    Affine registration of face landmarks to neutral face.

    Args:
        face_lms(array): face landmarks to register with shape (n,136).
                         Columns 0~67 are x coordinates and 68~136 are y coordinates
        neutral(array): target neutral face array that face_lm will be registered
        method(str or list): If string, register to all landmarks ('fullface', default),
                    or inner parts of face nose,mouth,eyes, and brows ('inner').
                    If list, pass landmarks to register to e.g. [27, 28, 29, 30, 36, 39, 42, 45]
    Return:
        registered_lms: registered landmarks in shape (n,136)
    """
    assert type(face_lms) == np.ndarray, TypeError("face_lms must be type np.ndarray")
    assert face_lms.ndim == 2, ValueError("face_lms must be shape (n, 136)")
    assert face_lms.shape[1] == 136, ValueError("Must have 136 landmarks")
    registered_lms = []
    for row in face_lms:
        face = [row[:68], row[68:]]
        face = np.array(face).T
        #   Rotate face
        primary = np.array(face)
        secondary = np.array(neutral)
        n = primary.shape[0]
        pad = lambda x: np.hstack([x, np.ones((x.shape[0], 1))])
        unpad = lambda x: x[:, :-1]
        X1, Y1 = pad(primary), pad(secondary)
        if type(method) == str:
            if method == "fullface":
                A, res, rank, s = np.linalg.lstsq(X1, Y1, rcond=None)
            elif method == "inner":
                A, res, rank, s = np.linalg.lstsq(X1[17:, :], Y1[17:, :], rcond=None)
            else:
                raise ValueError("method is either 'fullface' or 'inner'")
        elif type(method) == list:
            A, res, rank, s = np.linalg.lstsq(X1[method], Y1[method], rcond=None)
        else:
            raise TypeError(
                "method is string ('fullface','inner') or list of landmarks"
            )
        transform = lambda x: unpad(np.dot(pad(x), A))
        registered_lms.append(transform(primary).T.reshape(1, 136).ravel())
    return np.array(registered_lms)


def convert68to49(points):
    """
    Function slightly modified from https://github.com/D-X-Y/landmark-detection/blob/7bc7a5dbdbda314653124a4596f3feaf071e8589/SAN/lib/datasets/dataset_utils.py#L169
    to fit pytorch tensors.
    Converts 68 point landmarks to 49 point landmarks
    Args:
        points: landmark points of shape (2,68) or (3,68)
    Return:
        cpoints: converted 49 landmark points of shape (2,49)
    """
    assert (
        len(points.shape) == 2
        and (points.shape[0] == 3 or points.shape[0] == 2)
        and points.shape[1] == 68
    ), "The shape of points is not right : {}".format(points.shape)

    if isinstance(points, torch.Tensor):
        points = points.clone()
        out = torch.ones((68,), dtype=torch.bool)
    elif type(points) is np.ndarray:
        points = points.copy()
        out = np.ones((68,)).astype("bool")

    out[[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 60, 64]] = False
    cpoints = points[:, out]

    assert len(cpoints.shape) == 2 and cpoints.shape[1] == 49
    return cpoints


class BBox(object):
    # https://github.com/cunjian/pytorch_face_landmark/
    # bbox is a list of [left, right, top, bottom]
    def __init__(self, bbox):
        self.left = bbox[0]
        self.right = bbox[1]
        self.top = bbox[2]
        self.bottom = bbox[3]
        self.x = bbox[0]
        self.y = bbox[2]
        self.w = bbox[1] - bbox[0]
        self.h = bbox[3] - bbox[2]

    # scale to [0,1]
    def projectLandmark(self, landmark):
        landmark_ = np.asarray(np.zeros(landmark.shape))
        for i, point in enumerate(landmark):
            landmark_[i] = ((point[0] - self.x) / self.w, (point[1] - self.y) / self.h)
        return landmark_

    # landmark of (5L, 2L) from [0,1] to real range
    def reprojectLandmark(self, landmark):
        landmark_ = np.asarray(np.zeros(landmark.shape))
        for i, point in enumerate(landmark):
            x = point[0] * self.w + self.x
            y = point[1] * self.h + self.y
            landmark_[i] = (x, y)
        return landmark_


def drawLandmark(img, bbox, landmark):
    # https://github.com/cunjian/pytorch_face_landmark/
    """
    Input:
    - img: gray or RGB
    - bbox: type of BBox
    - landmark: reproject landmark of (5L, 2L)
    Output:
    - img marked with landmark and bbox
    """
    img_ = img.copy()
    cv2.rectangle(
        img_, (bbox.left, bbox.top), (bbox.right, bbox.bottom), (0, 0, 255), 2
    )
    for x, y in landmark:
        cv2.circle(img_, (int(x), int(y)), 3, (0, 255, 0), -1)
    return img_


def drawLandmark_multiple(img, bbox, landmark):
    # https://github.com/cunjian/pytorch_face_landmark/
    """
    Input:
    - img: gray or RGB
    - bbox: type of BBox
    - landmark: reproject landmark of (5L, 2L)
    Output:
    - img marked with landmark and bbox
    """
    cv2.rectangle(img, (bbox.left, bbox.top), (bbox.right, bbox.bottom), (0, 0, 255), 2)
    for x, y in landmark:
        cv2.circle(img, (int(x), int(y)), 2, (0, 255, 0), -1)
    return img



import matplotlib.pyplot as plt
import numpy as np
from os.path import join
from .utils import get_test_data_path
from feat import Fex, Detector
from feat.utils import read_openface, read_affectiva
from feat.plotting import plot_face, draw_lineface, draw_vectorfield, predict
import matplotlib
import pytest

# matplotlib.use('TkAgg')
matplotlib.use("Agg")


def assert_plot_shape(ax):
    assert ax.get_ylim() == (-240.0, -50.0)
    assert ax.get_xlim() == (25.0, 172.0)


feature_length = 20
au = np.ones(feature_length)
au2 = np.ones(feature_length) * 3


def testpredict():
    landmarks = predict(au)
    assert landmarks.shape == (2, 68)
    with pytest.raises(ValueError):
        predict(au, model=[0])
    with pytest.raises(ValueError):
        predict(au[:-1])


def test_draw_lineface():
    landmarks = predict(au)
    draw_lineface(currx=landmarks[0, :], curry=landmarks[1, :])
    assert_plot_shape(plt.gca())
    plt.close()


def test_draw_vectorfield():
    draw_vectorfield(reference=predict(au), target=predict(au=au2))
    assert_plot_shape(plt.gca())
    plt.close()
    with pytest.raises(ValueError):
        draw_vectorfield(
            reference=predict(au).reshape(4, 2 * feature_length), target=predict(au=au2)
        )
    with pytest.raises(ValueError):
        draw_vectorfield(
            reference=predict(au), target=predict(au=au2).reshape(4, 2 * feature_length)
        )


def test_plot_face():
    # test plotting method
    fx = Fex(
        filename=join(get_test_data_path(), "iMotions_Test_v6.txt"),
        sampling_freq=30,
        detector="FACET",
    )
    fx = fx.read_file()
    ax = fx.plot_aus(row_n=0)
    assert_plot_shape(ax)
    plt.close()

    fx = Fex(
        filename=join(get_test_data_path(), "OpenFace_Test.csv"),
        sampling_freq=30,
        detector="OpenFace",
    )
    fx = fx.read_file()
    ax = fx.plot_aus(row_n=0)
    assert_plot_shape(ax)
    plt.close()

    fx = Fex(
        filename=join(get_test_data_path(), "sample_affectiva-api-app_output.json"),
        sampling_freq=30,
        detector="Affectiva",
    )
    fx = fx.read_file(orig_cols=False)
    ax = fx.plot_aus(row_n=0)
    assert_plot_shape(ax)
    plt.close()

    # test plot in util
    plot_face()
    assert_plot_shape(plt.gca())
    plt.close()

    plot_face(au=au, vectorfield={"reference": predict(au2)})
    assert_plot_shape(plt.gca())
    plt.close()

    with pytest.raises(ValueError):
        plot_face(model=au, au=au, vectorfield={"reference": predict(au2)})
    with pytest.raises(ValueError):
        plot_face(model=au, au=au, vectorfield=[])
    with pytest.raises(ValueError):
        plot_face(model=au, au=au, vectorfield={"noreference": predict(au2)})


def test_plot_muscle():
    test_file = join(get_test_data_path(), "OpenFace_Test.csv")
    _, ax = plt.subplots(figsize=(4, 5))
    openface = read_openface(test_file)
    ax = openface.plot_aus(12, ax=ax, muscles={"all": "heatmap"}, gaze=None)
    assert_plot_shape(plt.gca())
    plt.close()


def test_plot_detections():
    test_data_dir = get_test_data_path()
    test_image = join(test_data_dir, "input.jpg")
    detector = Detector()
    image_prediction = detector.detect_image(test_image)
    axes = image_prediction.plot_detections()
    assert axes[1].get_xlim() == (0.0, 1.1)
    plt.close()

    axes = image_prediction.plot_detections(muscle=True)
    assert axes[1].get_xlim() == (0.0, 1.1)
    plt.close()

    image_prediction2 = image_prediction.copy()
    image_prediction2["input"] = "NO_SUCH_FILE_EXISTS"
    axes = image_prediction2.plot_detections()
    assert axes[1].get_xlim() == (0.0, 1.1)
    plt.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `feat` package."""

from feat.detector import Detector
from feat.data import Fex
from feat.utils import get_resource_path
from feat.tests.utils import get_test_data_path
import pandas as pd
import feat
import os
import wget
import cv2
import numpy as np

inputFname = os.path.join(get_test_data_path(), "sampler0000.jpg")
img01 = cv2.imread(inputFname)
h, w, _ = img01.shape


def test_detector():
    detector = Detector(n_jobs=1)
    assert detector["n_jobs"] == 1
    assert type(detector) == Detector


def test_faceboxes():
    # Face Detector Test Case:
    detector01 = Detector(
        face_model="FaceBoxes",
        landmark_model=None,
        au_model=None,
        emotion_model=None,
        n_jobs=1,
    )
    out = detector01.detect_faces(img01)
    bbox_left = out[0][0]
    assert bbox_left != None
    bbox_right = out[0][1]
    bbox_top = out[0][2]
    bbox_bottom = out[0][3]
    assert len(out[0]) == 5
    assert (
        bbox_left > 0
        and bbox_right > 0
        and bbox_top > 0
        and bbox_bottom > 0
        and bbox_left < bbox_right
        and bbox_top < bbox_bottom
        and bbox_left < w
        and bbox_right < w
        and bbox_top < h
        and bbox_bottom < h
    )


def test_retinaface():
    detector02 = Detector(
        face_model="RetinaFace",
        landmark_model=None,
        au_model=None,
        emotion_model=None,
        n_jobs=1,
    )
    out = detector02.detect_faces(img01)
    bbox_left = out[0][0]
    assert bbox_left != None
    bbox_right = out[0][1]
    bbox_top = out[0][2]
    bbox_bottom = out[0][3]
    assert len(out[0]) == 5
    assert (
        bbox_left > 0
        and bbox_right > 0
        and bbox_top > 0
        and bbox_bottom > 0
        and bbox_left < bbox_right
        and bbox_top < bbox_bottom
        and bbox_left < w
        and bbox_right < w
        and bbox_top < h
        and bbox_bottom < h
    )


def test_mtcnn():
    detector03 = Detector(
        face_model="MTCNN",
        landmark_model=None,
        au_model=None,
        emotion_model=None,
        n_jobs=1,
    )
    out = detector03.detect_faces(img01)
    bbox_left = out[0][0]
    assert bbox_left != None
    bbox_right = out[0][1]
    bbox_top = out[0][2]
    bbox_bottom = out[0][3]
    assert len(out[0]) == 5
    assert (
        bbox_left > 0
        and bbox_right > 0
        and bbox_top > 0
        and bbox_bottom > 0
        and bbox_left < bbox_right
        and bbox_top < bbox_bottom
        and bbox_left < w
        and bbox_right < w
        and bbox_top < h
        and bbox_bottom < h
    )


def test_mobilefacenet():
    # Landmark Detector Test Case:
    detector01 = Detector(
        face_model="RetinaFace", emotion_model=None, landmark_model="MobileFaceNet"
    )
    bboxes = detector01.detect_faces(img01)
    landmarks = detector01.detect_landmarks(img01, bboxes)
    assert landmarks[0].shape == (68, 2)
    assert (
        np.any(landmarks[0][:, 0] > 0)
        and np.any(landmarks[0][:, 0] < w)
        and np.any(landmarks[0][:, 1] > 0)
        and np.any(landmarks[0][:, 1] < h)
    )


def test_mobilenet():
    detector02 = Detector(
        face_model="RetinaFace", emotion_model=None, landmark_model="MobileNet"
    )
    bboxes = detector02.detect_faces(img01)
    landmarks = detector02.detect_landmarks(img01, bboxes)
    assert landmarks[0].shape == (68, 2)
    assert (
        np.any(landmarks[0][:, 0] > 0)
        and np.any(landmarks[0][:, 0] < w)
        and np.any(landmarks[0][:, 1] > 0)
        and np.any(landmarks[0][:, 1] < h)
    )


def test_pfld():
    detector03 = Detector(
        face_model="RetinaFace", emotion_model=None, landmark_model="PFLD"
    )
    bboxes = detector03.detect_faces(img01)
    landmarks = detector03.detect_landmarks(img01, bboxes)
    assert landmarks[0].shape == (68, 2)
    assert (
        np.any(landmarks[0][:, 0] > 0)
        and np.any(landmarks[0][:, 0] < w)
        and np.any(landmarks[0][:, 1] > 0)
        and np.any(landmarks[0][:, 1] < h)
    )


def test_jaanet():
    # AU Detection Case:
    detector1 = Detector(
        face_model="RetinaFace",
        emotion_model=None,
        landmark_model="MobileFaceNet",
        au_model="jaanet",
    )
    bboxes = detector1.detect_faces(img01)
    lands = detector1.detect_landmarks(img01, bboxes)
    aus = detector1.detect_aus(img01, lands)
    assert np.sum(np.isnan(aus)) == 0
    assert aus.shape[-1] == 12


def test_drml():
    # AU Detection Case2:
    inputFname = os.path.join(get_test_data_path(), "sampler0000.jpg")
    img01 = cv2.imread(inputFname)
    detector1 = Detector(
        face_model="RetinaFace",
        emotion_model=None,
        landmark_model="MobileFaceNet",
        au_model="drml",
    )
    bboxes = detector1.detect_faces(img01)
    lands = detector1.detect_landmarks(img01, bboxes)
    aus = detector1.detect_aus(img01, lands)
    assert np.sum(np.isnan(aus)) == 0
    assert aus.shape[-1] == 12


def test_resmasknet():
    inputFname = os.path.join(get_test_data_path(), "sampler0000.jpg")
    detector1 = Detector(emotion_model="resmasknet")
    out = detector1.detect_image(inputFname)
    assert out.emotions()["neutral"].values > 0.5


def test_detect_image():
    # Test detect image
    detector = Detector(n_jobs=1)
    inputFname = os.path.join(get_test_data_path(), "input.jpg")
    out = detector.detect_image(inputFname=inputFname)
    assert type(out) == Fex
    assert len(out) == 1
    assert out.happiness.values[0] > 0

    outputFname = os.path.join(get_test_data_path(), "output.csv")
    out = detector.detect_image(inputFname=inputFname, outputFname=outputFname)
    assert out
    assert os.path.exists(outputFname)
    out = pd.read_csv(outputFname)
    assert out.happiness.values[0] > 0


def test_multiface():
    inputFname2 = os.path.join(
        get_test_data_path(), "tim-mossholder-hOF1bWoet_Q-unsplash.jpg"
    )
    img02 = cv2.imread(inputFname2)
    detector = Detector(
        face_model="RetinaFace",
        emotion_model="fer",
        landmark_model="PFLD",
        au_model="jaanet",
    )
    files = detector.process_frame(img02, 0)
    assert files.shape[0] == 5


def test_detect_video():
    # Test detect video
    detector = Detector(n_jobs=1)
    inputFname = os.path.join(get_test_data_path(), "input.mp4")
    out = detector.detect_video(inputFname=inputFname, skip_frames=60)
    assert len(out) == 2


def test_detect_video_parallel():
    # Test detect video
    detector = Detector(n_jobs=2)
    inputFname = os.path.join(get_test_data_path(), "input.mp4")
    out = detector.detect_video(inputFname=inputFname, skip_frames=20, verbose=True)
    assert len(out) == 4

    outputFname = os.path.join(get_test_data_path(), "output.csv")
    out = detector.detect_video(
        inputFname=inputFname, outputFname=outputFname, skip_frames=10
    )
    assert out
    assert os.path.exists(outputFname)
    out = pd.read_csv(outputFname)
    assert out.happiness.values.max() > 0


def test_simultaneous():
    # Test processing everything:
    detector04 = Detector(
        face_model="RetinaFace",
        emotion_model="fer",
        landmark_model="PFLD",
        au_model="jaanet",
    )
    files = detector04.process_frame(img01, 0)
